import os

from packaging import version

# for using legacy Keras 2, and not Keras 3 installed by default from TensorFlow 2.16
os.environ["TF_USE_LEGACY_KERAS"] = "1"
os.environ["KERAS_BACKEND"] = "tensorflow"
# LF: do not remove this import. This ensures that CUDA 12 from tensorflow
#    get initialised before CUDA 11 from pytorch.
import tf_keras as keras  # noqa: F401

from delft import DELFT_PROJECT_DIR
from delft.utilities.misc import print_parameters

# ask tensorflow to be quiet and not print hundred lines of logs
from delft.utilities.Transformer import DEFAULT_TRANSFORMER_TOKENIZER_DIR, TRANSFORMER_CONFIG_FILE_NAME

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
os.environ["TOKENIZERS_PARALLELISM"] = "false"

import json
import math
import re
import time
import warnings
from itertools import islice

import numpy as np

warnings.filterwarnings("ignore", category=FutureWarning)
warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.filterwarnings("ignore", category=UserWarning)

import tensorflow as tf

tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)
tf.get_logger().setLevel("ERROR")

from tensorflow.python.util import deprecation

deprecation._PRINT_DEPRECATION_WARNINGS = False

# unfortunately when running in graph mode, we cannot use BERT pre-trained,
# see https://github.com/huggingface/transformers/issues/3086
# but this is apparently not useful anyway to disable eager mode here, because
# the Keras API compiles models before running them
# from tensorflow.python.framework.ops import disable_eager_execution
# disable_eager_execution()

import transformers

from delft.sequenceLabelling.config import ModelConfig, TrainingConfig
from delft.sequenceLabelling.evaluation import classification_report, get_report
from delft.sequenceLabelling.models import get_model
from delft.sequenceLabelling.preprocess import Preprocessor, prepare_preprocessor
from delft.sequenceLabelling.tagger import Tagger
from delft.sequenceLabelling.trainer import (
    CONFIG_FILE_NAME,
    DEFAULT_WEIGHT_FILE_NAME,
    PROCESSOR_FILE_NAME,
    Scorer,
    Trainer,
)
from delft.utilities.Embeddings import Embeddings, load_resource_registry
from delft.utilities.numpy import concatenate_or_none

transformers.logging.set_verbosity(transformers.logging.ERROR)


class Sequence(object):
    # number of parallel worker for the data generator
    nb_workers = 6

    def __init__(
        self,
        model_name=None,
        architecture=None,
        embeddings_name=None,
        char_emb_size=25,
        max_char_length=30,
        char_lstm_units=25,
        word_lstm_units=100,
        max_sequence_length=300,
        dropout=0.5,
        recurrent_dropout=0.25,
        batch_size=20,
        optimizer="adam",
        learning_rate=None,
        lr_decay=0.9,
        clip_gradients=5.0,
        max_epoch=50,
        early_stop=True,
        patience=5,
        max_checkpoints_to_keep=0,
        log_dir=None,
        fold_number=1,
        multiprocessing=True,
        num_workers=1,
        features_indices=None,
        transformer_name: str = None,
        report_to_wandb=False,
        nb_workers=6,
    ):

        if model_name is None:
            # add a dummy name based on the architecture
            model_name = architecture
            if embeddings_name is not None:
                model_name += "_" + embeddings_name
            if transformer_name is not None:
                model_name += "_" + transformer_name

        self.model = None
        self.models = None
        self.p: Preprocessor = None
        self.log_dir = log_dir
        self.embeddings_name = embeddings_name

        word_emb_size = 0
        self.nb_workers = nb_workers
        self.embeddings = None
        self.model_local_path = None

        self.report_to_wandb = report_to_wandb

        self.registry = load_resource_registry(os.path.join(DELFT_PROJECT_DIR, "resources-registry.json"))

        if self.embeddings_name is not None:
            self.embeddings = Embeddings(self.embeddings_name, resource_registry=self.registry)
            word_emb_size = self.embeddings.embed_size
        else:
            self.embeddings = None

        if learning_rate is None:
            if transformer_name is None:
                learning_rate = 0.001
            else:
                learning_rate = 2e-5

        self.model_config = ModelConfig(
            model_name=model_name,
            architecture=architecture,
            embeddings_name=embeddings_name,
            word_embedding_size=word_emb_size,
            char_emb_size=char_emb_size,
            char_lstm_units=char_lstm_units,
            max_char_length=max_char_length,
            word_lstm_units=word_lstm_units,
            max_sequence_length=max_sequence_length,
            dropout=dropout,
            recurrent_dropout=recurrent_dropout,
            fold_number=fold_number,
            batch_size=batch_size,
            features_indices=features_indices,
            transformer_name=transformer_name,
        )

        self.training_config = TrainingConfig(
            learning_rate,
            batch_size,
            optimizer,
            lr_decay,
            clip_gradients,
            max_epoch,
            early_stop,
            patience,
            max_checkpoints_to_keep,
            multiprocessing,
            num_workers,
        )

        if report_to_wandb:
            import wandb
            from dotenv import load_dotenv

            load_dotenv(override=True)
            if os.getenv("WANDB_API_KEY") is None:
                print("Warning: WANDB_API_KEY not set, wandb disabled")
                self.report_to_wandb = False
                return
            wandb.init(
                name=model_name,
                config={
                    "model_name": self.model_config.model_name,
                    "architecture": self.model_config.architecture,
                    "transformer_name": self.model_config.transformer_name,
                    "embeddings_name": self.model_config.embeddings_name,
                    "embedding_size": self.model_config.word_embedding_size,
                    "batch_size": self.training_config.batch_size,
                    "learning_rate": self.training_config.learning_rate,
                    "lr_decay": self.training_config.lr_decay,
                    "max_epoch": self.training_config.max_epoch,
                    "early_stop": self.training_config.early_stop,
                    "patience": self.training_config.patience,
                    "char_emb_size": self.model_config.char_embedding_size,
                    "max_char_length": self.model_config.max_char_length,
                    "max_sequence_length": self.model_config.max_sequence_length,
                    "dropout": self.model_config.dropout,
                    "recurrent_dropout": self.model_config.recurrent_dropout,
                    "optimizer": self.training_config.optimizer,
                    "clip_gradients": self.training_config.clip_gradients,
                },
            )

            wandb.define_metric("f1", summary="max")

    def train(
        self,
        x_train,
        y_train,
        f_train=None,
        x_valid=None,
        y_valid=None,
        f_valid=None,
        incremental=False,
        callbacks=None,
        multi_gpu=False,
    ):
        if self.report_to_wandb:
            from wandb.integration.keras import WandbMetricsLogger

            callbacks = (
                callbacks
                + [
                    WandbMetricsLogger(),
                    # WandbModelCheckpoint("models", monitor='f1', mode='max')
                ]
                if callbacks is not None
                else [
                    WandbMetricsLogger(),
                    # WandbModelCheckpoint("models", monitor='f1', mode='max')
                ]
            )

        # TBD if valid is None, segment train to get one if early_stop is True
        if multi_gpu:
            strategy = tf.distribute.MirroredStrategy()
            print("Running with multi-gpu. Number of devices: {}".format(strategy.num_replicas_in_sync))

            # This trick avoid an exception being through when the --multi-gpu approach is used on a single GPU system.
            # It might be removed with TF 2.10 https://github.com/tensorflow/tensorflow/issues/50487
            import atexit

            atexit.register(strategy._extended._collective_ops._pool.close)  # type: ignore

            with strategy.scope():
                self.train_(x_train, y_train, f_train, x_valid, y_valid, f_valid, incremental, callbacks)
        else:
            self.train_(x_train, y_train, f_train, x_valid, y_valid, f_valid, incremental, callbacks)

    def train_(
        self,
        x_train,
        y_train,
        f_train=None,
        x_valid=None,
        y_valid=None,
        f_valid=None,
        incremental=False,
        callbacks=None,
    ):
        # TBD if valid is None, segment train to get one if early_stop is True

        # we concatenate all the training+validation data to create the model vocabulary
        if x_valid is not None:
            x_all = np.concatenate((x_train, x_valid), axis=0)
        else:
            x_all = x_train

        if y_valid is not None:
            y_all = np.concatenate((y_train, y_valid), axis=0)
        else:
            y_all = y_train

        features_all = concatenate_or_none((f_train, f_valid), axis=0)

        if incremental:
            if self.model is None and self.models is None:
                print("error: you must load a model first for an incremental training")
                return
            print("Incremental training from loaded model", self.model_config.model_name)
            # update the preprocessor for the new chars and labels
            self.p.extend(x_all, y_all)
        else:
            # init a new "fresh" model
            self.p = prepare_preprocessor(x_all, y_all, features=features_all, model_config=self.model_config)

            self.model_config.char_vocab_size = len(self.p.vocab_char)
            self.model_config.case_vocab_size = len(self.p.vocab_case)

            self.model = get_model(self.model_config, self.p, len(self.p.vocab_tag), load_pretrained_weights=True)

        print_parameters(self.model_config, self.training_config)
        self.model.print_summary()

        # uncomment to plot graph
        # from tensorflow.keras.utils import plot_model
        # plot_model(self.model,
        #    to_file='data/models/textClassification/'+self.model_config.model_name+'_'+self.model_config.architecture+'.png')

        trainer = Trainer(
            self.model,
            self.models,
            self.embeddings,
            self.model_config,
            self.training_config,
            checkpoint_path=self.log_dir,
            preprocessor=self.p,
            transformer_preprocessor=self.model.transformer_preprocessor,
            enable_wandb=self.report_to_wandb,
            nb_workers=self.nb_workers,
        )
        trainer.train(
            x_train, y_train, x_valid, y_valid, features_train=f_train, features_valid=f_valid, callbacks=callbacks
        )

    def train_nfold(
        self,
        x_train,
        y_train,
        x_valid=None,
        y_valid=None,
        f_train=None,
        f_valid=None,
        incremental=False,
        callbacks=None,
        multi_gpu=False,
    ):
        if multi_gpu:
            strategy = tf.distribute.MirroredStrategy()
            print("Running with multi-gpu. Number of devices: {}".format(strategy.num_replicas_in_sync))

            # This trick avoid an exception being through when the --multi-gpu approach is used on a single GPU system.
            # It might be removed with TF 2.10 https://github.com/tensorflow/tensorflow/issues/50487
            import atexit

            atexit.register(strategy._extended._collective_ops._pool.close)  # type: ignore

            with strategy.scope():
                self.train_nfold_(x_train, y_train, x_valid, y_valid, f_train, f_valid, incremental, callbacks)
        else:
            self.train_nfold_(x_train, y_train, x_valid, y_valid, f_train, f_valid, incremental, callbacks)

    def train_nfold_(
        self,
        x_train,
        y_train,
        x_valid=None,
        y_valid=None,
        f_train=None,
        f_valid=None,
        incremental=False,
        callbacks=None,
    ):
        x_all = np.concatenate((x_train, x_valid), axis=0) if x_valid is not None else x_train
        y_all = np.concatenate((y_train, y_valid), axis=0) if y_valid is not None else y_train
        features_all = concatenate_or_none((f_train, f_valid), axis=0)

        if incremental:
            if self.model is None and self.models is None:
                print("error: you must load a model first for an incremental training")
                return

            print("Incremental training from loaded model", self.model_config.model_name)
            self.model.print_summary()
        else:
            self.p = prepare_preprocessor(x_all, y_all, features=features_all, model_config=self.model_config)
            self.model_config.char_vocab_size = len(self.p.vocab_char)
            self.model_config.case_vocab_size = len(self.p.vocab_case)
            self.models = []

        trainer = Trainer(
            self.model,
            self.models,
            self.embeddings,
            self.model_config,
            self.training_config,
            checkpoint_path=self.log_dir,
            preprocessor=self.p,
            nb_workers=self.nb_workers,
        )

        trainer.train_nfold(x_train, y_train, x_valid, y_valid, f_train=f_train, f_valid=f_valid, callbacks=callbacks)

    def eval(self, x_test, y_test, features=None):
        if self.model_config.fold_number > 1:
            self.eval_nfold(x_test, y_test, features=features)
        else:
            self.eval_single(x_test, y_test, features=features)

    def eval_single(self, x_test, y_test, features=None):
        if self.model is None:
            raise (OSError("Could not find a model."))
        print_parameters(self.model_config, self.training_config)
        self.model.print_summary()

        if self.model_config.transformer_name is None:
            # we can use a data generator for evaluation

            # Prepare test data(steps, generator)
            generator = self.model.get_generator()
            test_generator = generator(
                x_test,
                y_test,
                batch_size=self.model_config.batch_size,
                preprocessor=self.p,
                char_embed_size=self.model_config.char_embedding_size,
                max_sequence_length=self.model_config.max_sequence_length,
                embeddings=self.embeddings,
                shuffle=False,
                features=features,
                output_input_offsets=True,
                use_chain_crf=self.model_config.use_chain_crf,
            )

            # Build the evaluator and evaluate the model
            scorer = Scorer(
                test_generator,
                self.p,
                evaluation=True,
                use_crf=self.model_config.use_crf,
                use_chain_crf=self.model_config.use_chain_crf,
            )
            scorer.model = self.model
            scorer.on_epoch_end(epoch=-1)
        else:
            # the architecture model uses a transformer layer
            # note that we could also use the above test_generator, but as an alternative here we check the
            # test/prediction alignment of tokens and the validity of the maximum sequence input length
            # wrt the length of the test sequences

            tagger = Tagger(
                self.model,
                self.model_config,
                self.embeddings,
                preprocessor=self.p,
                transformer_preprocessor=self.model.transformer_preprocessor,
            )
            y_pred_pairs = tagger.tag(x_test, output_format=None, features=features)

            # keep only labels
            y_pred = []
            for result in y_pred_pairs:
                result_labels = []
                for pair in result:
                    result_labels.append(pair[1])
                y_pred.append(result_labels)

            nb_alignment_issues = 0
            for i in range(len(y_test)):
                if len(y_test[i]) != len(y_pred[i]):
                    # print("y_test:", y_test[i])
                    # print("y_pred:", y_pred[i])

                    nb_alignment_issues += 1
                    # BERT tokenizer appears to introduce some additional tokens without ## prefix,
                    # but we normally handled that well when predicting.
                    # To be very conservative, the following ensure the number of tokens always
                    # match, but it should never be used in practice.
                    if len(y_test[i]) < len(y_pred[i]):
                        y_test[i] = y_test[i] + ["O"] * (len(y_pred[i]) - len(y_test[i]))
                    if len(y_test[i]) > len(y_pred[i]):
                        y_pred[i] = y_pred[i] + ["O"] * (len(y_test[i]) - len(y_pred[i]))

            if nb_alignment_issues > 0:
                print("number of alignment issues with test set:", nb_alignment_issues)
                print("to solve them consider increasing the maximum sequence input length of the model and retrain")

            report, report_as_map = classification_report(y_test, y_pred, digits=4)
            print(report)

    def eval_nfold(self, x_test, y_test, features=None):
        if self.models is not None:
            total_f1 = 0
            best_f1 = 0
            best_index = 0
            worst_f1 = 1
            worst_index = 0
            reports = []
            reports_as_map = []
            total_precision = 0
            total_recall = 0
            for i in range(self.model_config.fold_number):
                if self.model_config.transformer_name is None:
                    the_model = self.models[i]
                    bert_preprocessor = None
                else:
                    # the architecture model uses a transformer layer, it is large and needs to be loaded from disk
                    dir_path = "data/models/sequenceLabelling/"
                    weight_file = DEFAULT_WEIGHT_FILE_NAME.replace(".hdf5", str(i) + ".hdf5")
                    self.model = get_model(
                        self.model_config,
                        self.p,
                        ntags=len(self.p.vocab_tag),
                        load_pretrained_weights=False,
                        local_path=os.path.join(dir_path, self.model_config.model_name),
                    )
                    self.model.load(filepath=os.path.join(dir_path, self.model_config.model_name, weight_file))
                    the_model = self.model
                    bert_preprocessor = self.model.transformer_preprocessor

                if i == 0:
                    the_model.print_summary()
                    print_parameters(self.model_config, self.training_config)

                print("\n------------------------ fold " + str(i) + " --------------------------------------")

                # we can use a data generator for evaluation
                # Prepare test data(steps, generator)
                generator = the_model.get_generator()
                test_generator = generator(
                    x_test,
                    y_test,
                    batch_size=self.model_config.batch_size,
                    preprocessor=self.p,
                    bert_preprocessor=bert_preprocessor,
                    char_embed_size=self.model_config.char_embedding_size,
                    max_sequence_length=self.model_config.max_sequence_length,
                    embeddings=self.embeddings,
                    shuffle=False,
                    features=features,
                    output_input_offsets=True,
                    use_chain_crf=self.model_config.use_chain_crf,
                )

                # Build the evaluator and evaluate the model
                scorer = Scorer(
                    test_generator,
                    self.p,
                    evaluation=True,
                    use_crf=self.model_config.use_crf,
                    use_chain_crf=self.model_config.use_chain_crf,
                )
                scorer.model = the_model
                scorer.on_epoch_end(epoch=-1)
                f1 = scorer.f1
                precision = scorer.precision
                recall = scorer.recall
                reports.append(scorer.report)
                reports_as_map.append(scorer.report_as_map)

                if best_f1 < f1:
                    best_f1 = f1
                    best_index = i
                if worst_f1 > f1:
                    worst_f1 = f1
                    worst_index = i
                total_f1 += f1
                total_precision += precision
                total_recall += recall

            fold_average_evaluation = {"labels": {}, "micro": {}, "macro": {}}

            micro_f1 = total_f1 / self.model_config.fold_number
            micro_precision = total_precision / self.model_config.fold_number
            micro_recall = total_recall / self.model_config.fold_number

            micro_eval_block = {"f1": micro_f1, "precision": micro_precision, "recall": micro_recall}
            fold_average_evaluation["micro"] = micro_eval_block

            # field-level average over the n folds
            labels = []
            for label in sorted(self.p.vocab_tag):
                if label == "O" or label == "<PAD>":
                    continue
                if label.startswith("B-") or label.startswith("S-") or label.startswith("I-") or label.startswith("E-"):
                    label = label[2:]

                if label in labels:
                    continue
                labels.append(label)

                sum_p = 0
                sum_r = 0
                sum_f1 = 0
                sum_support = 0
                for j in range(0, self.model_config.fold_number):
                    if label not in reports_as_map[j]["labels"]:
                        continue
                    report_as_map = reports_as_map[j]["labels"][label]
                    sum_p += report_as_map["precision"]
                    sum_r += report_as_map["recall"]
                    sum_f1 += report_as_map["f1"]
                    sum_support += report_as_map["support"]

                avg_p = sum_p / self.model_config.fold_number
                avg_r = sum_r / self.model_config.fold_number
                avg_f1 = sum_f1 / self.model_config.fold_number
                avg_support = sum_support / self.model_config.fold_number
                avg_support_dec = str(avg_support - int(avg_support))[1:]
                if avg_support_dec != "0":
                    avg_support = math.floor(avg_support)

                block_label = {"precision": avg_p, "recall": avg_r, "support": avg_support, "f1": avg_f1}
                fold_average_evaluation["labels"][label] = block_label

            print("----------------------------------------------------------------------")
            print("\n** Worst ** model scores - run", str(worst_index))
            print(reports[worst_index])

            print("\n** Best ** model scores - run", str(best_index))
            print(reports[best_index])

            fold_nb = self.model_config.fold_number
            self.model_config.fold_number = 1
            if self.model_config.transformer_name is None:
                self.model = self.models[best_index]
            else:
                dir_path = "data/models/sequenceLabelling/"
                weight_file = DEFAULT_WEIGHT_FILE_NAME.replace(".hdf5", str(best_index) + ".hdf5")
                # saved config file must be updated to single fold
                self.model.load(filepath=os.path.join(dir_path, self.model_config.model_name, weight_file))

            print("----------------------------------------------------------------------")
            print("\nAverage over", str(int(fold_nb)), "folds")
            print(get_report(fold_average_evaluation, digits=4, include_avgs=["micro"]))

    def tag(self, texts, output_format, features=None, batch_size=None, multi_gpu=False):
        if multi_gpu:
            strategy = tf.distribute.MirroredStrategy()
            print("Running with multi-gpu. Number of devices: {}".format(strategy.num_replicas_in_sync))

            # This trick avoid an exception being through when the --multi-gpu approach is used on a single GPU system.
            # It might be removed with TF 2.10 https://github.com/tensorflow/tensorflow/issues/50487
            if version.parse(tf.__version__) < version.parse("2.10.0"):
                import atexit

                atexit.register(strategy._extended._collective_ops._pool.close)  # type: ignore

            with strategy.scope():
                return self.tag_(texts, output_format, features, batch_size)
        else:
            return self.tag_(texts, output_format, features, batch_size)

    def tag_(self, texts, output_format, features=None, batch_size=None):
        # annotate a list of sentences, return the list of annotations in the
        # specified output_format

        if batch_size is not None:
            self.model_config.batch_size = batch_size
            print("---")
            print("batch_size (prediction):", self.model_config.batch_size)
            print("---")

        if self.model:
            tagger = Tagger(
                self.model,
                self.model_config,
                self.embeddings,
                preprocessor=self.p,
                transformer_preprocessor=self.model.transformer_preprocessor,
            )
            start_time = time.time()
            annotations = tagger.tag(texts, output_format, features=features)
            runtime = round(time.time() - start_time, 3)
            if output_format == "json":
                annotations["runtime"] = runtime
            # else:
            #    print("runtime: %s seconds " % (runtime))
            return annotations
        else:
            raise (OSError("Could not find a model." + str(self.model)))

    def tag_file(self, file_in, output_format, file_out, batch_size=None, multi_gpu=False):
        # Annotate a text file containing one sentence per line, the annotations are
        # written in the output file if not None, in the standard output otherwise.
        # Processing is streamed by batches so that we can process huge files without
        # memory issues

        if batch_size is not None:
            self.model_config.batch_size = batch_size
            print("---")
            print("batch_size (prediction):", self.model_config.batch_size)
            print("---")

        if self.model:
            tagger = Tagger(
                self.model,
                self.model_config,
                self.embeddings,
                preprocessor=self.p,
                transformer_preprocessor=self.model.transformer_preprocessor,
            )
            start_time = time.time()
            if file_out is not None:
                out = open(file_out, "w")
            first = True
            with open(file_in, "r") as f:
                texts = None
                while texts is None or len(texts) == self.model_config.batch_size * self.nb_workers:
                    texts = next_n_lines(f, self.model_config.batch_size * self.nb_workers)
                    annotations = tagger.tag(texts, output_format, multi_gpu=multi_gpu)
                    # if the following is true, we just output the JSON returned by the tagger without any modification
                    directDump = False
                    if first:
                        first = False
                        if len(texts) < self.model_config.batch_size * self.nb_workers:
                            runtime = round(time.time() - start_time, 3)
                            annotations["runtime"] = runtime
                            jsonString = json.dumps(annotations, sort_keys=False, indent=4, ensure_ascii=False)
                            if file_out is None:
                                print(jsonString)
                            else:
                                out.write(jsonString)
                            directDump = True
                        else:
                            # we need to modify a bit the JSON outputted by the tagger to glue the different batches
                            # output the general information attributes
                            jsonString = (
                                '{\n    "software": ' + json.dumps(annotations["software"], ensure_ascii=False) + ",\n"
                            )
                            jsonString += '    "date": ' + json.dumps(annotations["date"], ensure_ascii=False) + ",\n"
                            jsonString += '    "model": ' + json.dumps(annotations["model"], ensure_ascii=False) + ",\n"
                            jsonString += '    "texts": ['
                            if file_out is None:
                                print(jsonString, end="", flush=True)
                            else:
                                out.write(jsonString)
                            first = True
                            for jsonStr in annotations["texts"]:
                                jsonString = json.dumps(jsonStr, sort_keys=False, indent=4, ensure_ascii=False)
                                # jsonString = jsonString.replace('\n', '\n\t\t')
                                jsonString = re.sub("\n", "\n        ", jsonString)
                                if file_out is None:
                                    if not first:
                                        print(",\n        " + jsonString, end="", flush=True)
                                    else:
                                        first = False
                                        print("\n        " + jsonString, end="", flush=True)
                                else:
                                    if not first:
                                        out.write(",\n        ")
                                        out.write(jsonString)
                                    else:
                                        first = False
                                        out.write("\n        ")
                                        out.write(jsonString)
                    else:
                        for jsonStr in annotations["texts"]:
                            jsonString = json.dumps(jsonStr, sort_keys=False, indent=4, ensure_ascii=False)
                            jsonString = re.sub("\n", "\n        ", jsonString)
                            if file_out is None:
                                print(",\n        " + jsonString, end="", flush=True)
                            else:
                                out.write(",\n        ")
                                out.write(jsonString)

            runtime = round(time.time() - start_time, 3)
            if not directDump:
                jsonString = "\n    ],\n"
                jsonString += '    "runtime": ' + str(runtime)
                jsonString += "\n}\n"
                if file_out is None:
                    print(jsonString)
                else:
                    out.write(jsonString)

            if file_out is not None:
                out.close()
            # print("runtime: %s seconds " % (runtime))
        else:
            raise (OSError("Could not find a model."))

    def save(self, dir_path="data/models/sequenceLabelling/", weight_file=DEFAULT_WEIGHT_FILE_NAME):
        # create subfolder for the model if not already exists
        directory = os.path.join(dir_path, self.model_config.model_name)
        if not os.path.exists(directory):
            os.makedirs(directory)

        self.model_config.save(os.path.join(directory, CONFIG_FILE_NAME))
        print("model config file saved")

        self.p.save(os.path.join(directory, PROCESSOR_FILE_NAME))
        print("preprocessor saved")

        if self.model is None and self.model_config.fold_number > 1:
            print(
                "Error: model not saved. Evaluation need to be called first to select the best fold model to be saved"
            )
        else:
            self.model.save(os.path.join(directory, weight_file))

            # save pretrained transformer config if used in the model
            if self.model.transformer_config is not None:
                self.model.transformer_config.to_json_file(os.path.join(directory, TRANSFORMER_CONFIG_FILE_NAME))
                print("transformer config saved")

            if self.model.transformer_preprocessor is not None:
                self.model.transformer_preprocessor.tokenizer.save_pretrained(
                    os.path.join(directory, DEFAULT_TRANSFORMER_TOKENIZER_DIR)
                )
                print("transformer tokenizer saved")

        print("model saved")

    def load(self, dir_path="data/models/sequenceLabelling/", weight_file=DEFAULT_WEIGHT_FILE_NAME):
        model_path = os.path.join(dir_path, self.model_config.model_name)
        self.model_config = ModelConfig.load(os.path.join(model_path, CONFIG_FILE_NAME))

        if self.model_config.embeddings_name is not None:
            # load embeddings
            # Do not use cache in 'prediction/production' mode
            self.embeddings = Embeddings(
                self.model_config.embeddings_name, resource_registry=self.registry, use_cache=False
            )
            self.model_config.word_embedding_size = self.embeddings.embed_size
        else:
            self.embeddings = None
            self.model_config.word_embedding_size = 0

        self.p = Preprocessor.load(os.path.join(dir_path, self.model_config.model_name, PROCESSOR_FILE_NAME))
        self.model = get_model(
            self.model_config,
            self.p,
            ntags=len(self.p.vocab_tag),
            load_pretrained_weights=False,
            local_path=os.path.join(dir_path, self.model_config.model_name),
        )
        print("load weights from", os.path.join(dir_path, self.model_config.model_name, weight_file))
        self.model.load(filepath=os.path.join(dir_path, self.model_config.model_name, weight_file))
        self.model.print_summary()


def next_n_lines(file_opened, N):
    return [x.strip() for x in islice(file_opened, N)]



import logging

from delft.sequenceLabelling.tagger import get_entities_with_offsets
from delft.utilities.Tokenizer import tokenizeAndFilter

LOGGER = logging.getLogger(__name__)


def test_get_entities_with_offsets():
    original_string = "(Mo -x 1 T x ) 3 Sb 7 with \uf084 x 0.1"
    tokens = ["(", "Mo", "-", "x", "1", "T", "x", ")", "3", "Sb", "7", "with", "\uf084", "x", "0", ".", "1"]
    tags = [
        "B-<formula>",
        "I-<formula>",
        "I-<formula>",
        "I-<formula>",
        "I-<formula>",
        "I-<formula>",
        "I-<formula>",
        "I-<formula>",
        "I-<formula>",
        "I-<formula>",
        "I-<formula>",
        "O",
        "O",
        "B-<variable>",
        "B-<value>",
        "I-<value>",
        "I-<value>",
    ]
    [tag.split("-")[-1] for tag in tags]

    offsets = [
        (0, 1),
        (1, 3),
        (4, 5),
        (5, 6),
        (7, 8),
        (9, 10),
        (11, 12),
        (13, 14),
        (15, 16),
        (17, 19),
        (20, 21),
        (22, 26),
        (27, 28),
        (29, 30),
        (31, 32),
        (32, 33),
        (33, 34),
    ]

    [offsets[offsetIndex][1] != offsets[offsetIndex + 1][0] for offsetIndex in range(0, len(offsets) - 1)]

    for index in range(0, len(offsets)):
        chunk = original_string[offsets[index][0] : offsets[index][1]]

        assert chunk == tokens[index]

    entities_with_offsets = get_entities_with_offsets(tags, offsets)
    # (chunk_type, chunk_start, chunk_end, pos_start, pos_end)

    assert len(entities_with_offsets) == 3
    entity0 = entities_with_offsets[0]
    assert entity0[0] == "<formula>"
    entity_text = original_string[entity0[3] : entity0[4] + 1]
    assert entity_text == "(Mo -x 1 T x ) 3 Sb 7"
    assert tokens[entity0[1] : entity0[2]] == tokenizeAndFilter(entity_text)[0]

    entity1 = entities_with_offsets[1]
    assert entity1[0] == "<variable>"
    entity_text = original_string[entity1[3] : entity1[4] + 1]
    assert entity_text == "x"
    assert tokens[entity1[1] : entity1[2]] == tokenizeAndFilter(entity_text)[0]

    entity2 = entities_with_offsets[2]
    assert entity2[0] == "<value>"
    entity_text = original_string[entity2[3] : entity2[4] + 1]
    assert entity_text == "0.1"
    assert tokens[entity2[1] : entity2[2]] == tokenizeAndFilter(entity_text)[0]

    # for item in entities_with_offsets:
    #     type = item[0]
    #     token_start = item[1]
    #     token_end = item[2]
    #     char_start = item[3]
    #     char_end = item[4]
    #
    #     text = ''.join([tokens[idx] + (' ' if spaces[idx] else '') for idx in range(token_start, token_end)])
    #     if text.endswith(' '):
    #         text = text[0:-1]
    #
    #     assert text == original_string[char_start: char_end + 1]



import argparse
import json
import time

import numpy as np

from delft.textClassification import Classifier
from delft.textClassification.models import architectures
from delft.textClassification.reader import load_software_context_corpus_json, load_software_dataset_context_corpus_json
from delft.utilities.numpy import concatenate_or_none, shuffle_triple_with_view
from delft.utilities.Utilities import split_data_and_labels

"""
    A multiclass classifier to be used in combination with a software mention recognition model, for characterizing
    the nature of the mention of software in scientific and technical literature.
    This classifier predicts if the software introduced by a software mention in a sentence is likely:
    - used or not by the described work (class used)
    - a creation of the described work (class creation)
    - shared (class shared)

    For the software mention recognizer, see https://github.com/ourresearch/software-mentions
    and grobidTagger.py in the present project DeLFT.

    Best architecture/model is fine-tuned SciBERT.
"""

list_classes = ["used", "creation", "shared"]
class_weights = {0: 1.0, 1: 1.0, 2: 1.0}


def configure(architecture):
    batch_size = 256
    maxlen = 300
    patience = 5
    early_stop = True
    max_epoch = 60

    # default bert model parameters
    if architecture == "bert":
        batch_size = 32
        early_stop = False
        max_epoch = 6
        maxlen = 100

    return batch_size, maxlen, patience, early_stop, max_epoch


def train(embeddings_name, fold_count, architecture="gru", transformer=None):
    print("loading multiclass software context dataset...")
    xtr, y = load_software_context_corpus_json("data/textClassification/software/software-contexts.json.gz")
    xtr2, y2 = load_software_dataset_context_corpus_json(
        "data/textClassification/software/all_clean.classification_extra.json.gz"
    )

    xtr = concatenate_or_none([xtr, xtr2])
    y = concatenate_or_none([y, y2])

    xtr, y, _ = shuffle_triple_with_view(xtr, y)

    report_training_contexts(y)

    model_name = "software_context_" + architecture
    class_weights = None

    batch_size, maxlen, patience, early_stop, max_epoch = configure(architecture)

    model = Classifier(
        model_name,
        architecture=architecture,
        list_classes=list_classes,
        max_epoch=max_epoch,
        fold_number=fold_count,
        patience=patience,
        use_roc_auc=True,
        embeddings_name=embeddings_name,
        batch_size=batch_size,
        maxlen=maxlen,
        early_stop=early_stop,
        class_weights=class_weights,
        transformer_name=transformer,
    )

    if fold_count == 1:
        model.train(xtr, y)
    else:
        model.train_nfold(xtr, y)
    # saving the model
    model.save()


def train_and_eval(embeddings_name, fold_count, architecture="gru", transformer=None):
    print("loading multiclass software context dataset...")
    xtr, y = load_software_context_corpus_json("data/textClassification/software/software-contexts.json.gz")
    xtr2, y2 = load_software_dataset_context_corpus_json(
        "data/textClassification/software/all_clean.classification_extra.json.gz"
    )

    xtr = concatenate_or_none([xtr, xtr2])
    y = concatenate_or_none([y, y2])

    xtr, y, _ = shuffle_triple_with_view(xtr, y)

    report_training_contexts(y)

    model_name = "software_context_" + architecture
    class_weights = None

    # segment train and eval sets
    x_train, y_train, x_test, y_test = split_data_and_labels(xtr, y, 0.9)

    batch_size, maxlen, patience, early_stop, max_epoch = configure(architecture)

    print(list_classes)

    model = Classifier(
        model_name,
        architecture=architecture,
        list_classes=list_classes,
        max_epoch=max_epoch,
        fold_number=fold_count,
        patience=patience,
        use_roc_auc=True,
        embeddings_name=embeddings_name,
        batch_size=batch_size,
        maxlen=maxlen,
        early_stop=early_stop,
        class_weights=class_weights,
        transformer_name=transformer,
    )

    if fold_count == 1:
        model.train(x_train, y_train)
    else:
        model.train_nfold(x_train, y_train)
    model.eval(x_test, y_test)

    # saving the model
    model.save()


def train_binary(embeddings_name, fold_count, architecture="gru", transformer=None):
    print("loading multiclass software context dataset...")
    x_train, y_train = load_software_context_corpus_json("data/textClassification/software/software-contexts.json.gz")
    xtr2, y2 = load_software_dataset_context_corpus_json(
        "data/textClassification/software/all_clean.classification_extra.json.gz"
    )

    xtr = concatenate_or_none([x_train, xtr2])
    y = concatenate_or_none([y_train, y2])

    xtr, y, _ = shuffle_triple_with_view(xtr, y)

    report_training_contexts(y_train)

    for class_rank in range(len(list_classes)):
        model_name = "software_context_" + list_classes[class_rank] + "_" + architecture
        class_weights = None

        batch_size, maxlen, patience, early_stop, max_epoch = configure(architecture)

        y_train_class_rank = [[1, 0] if y[class_rank] == 1.0 else [0, 1] for y in y_train]
        y_train_class_rank = np.array(y_train_class_rank)

        list_classes_rank = [list_classes[class_rank], "not_" + list_classes[class_rank]]

        model = Classifier(
            model_name,
            architecture=architecture,
            list_classes=list_classes_rank,
            max_epoch=max_epoch,
            fold_number=fold_count,
            patience=patience,
            use_roc_auc=True,
            embeddings_name=embeddings_name,
            batch_size=batch_size,
            maxlen=maxlen,
            early_stop=early_stop,
            class_weights=class_weights,
            transformer_name=transformer,
        )

        if fold_count == 1:
            model.train(x_train, y_train_class_rank)
        else:
            model.train_nfold(x_train, y_train_class_rank)
        # saving the model
        model.save()


def train_and_eval_binary(embeddings_name, fold_count, architecture="gru", transformer=None):
    print("loading multiclass software context dataset...")
    xtr, y = load_software_context_corpus_json("data/textClassification/software/software-contexts.json.gz")
    xtr2, y2 = load_software_dataset_context_corpus_json(
        "data/textClassification/software/all_clean.classification_extra.json.gz"
    )

    xtr = concatenate_or_none([xtr, xtr2])
    y = concatenate_or_none([y, y2])

    xtr, y, _ = shuffle_triple_with_view(xtr, y)

    report_training_contexts(y)
    # segment train and eval sets
    x_train, y_train, x_test, y_test = split_data_and_labels(xtr, y, 0.9)

    for class_rank in range(len(list_classes)):
        model_name = "software_context_" + list_classes[class_rank] + "_" + architecture
        class_weights = None

        batch_size, maxlen, patience, early_stop, max_epoch = configure(architecture)

        y_train_class_rank = [[1, 0] if y[class_rank] == 1.0 else [0, 1] for y in y_train]
        y_test_class_rank = [[1, 0] if y[class_rank] == 1.0 else [0, 1] for y in y_test]

        y_train_class_rank = np.array(y_train_class_rank)
        y_test_class_rank = np.array(y_test_class_rank)

        list_classes_rank = [list_classes[class_rank], "not_" + list_classes[class_rank]]

        model = Classifier(
            model_name,
            architecture=architecture,
            list_classes=list_classes_rank,
            max_epoch=max_epoch,
            fold_number=fold_count,
            patience=patience,
            use_roc_auc=True,
            embeddings_name=embeddings_name,
            batch_size=batch_size,
            maxlen=maxlen,
            early_stop=early_stop,
            class_weights=class_weights,
            transformer_name=transformer,
        )

        if fold_count == 1:
            model.train(x_train, y_train_class_rank)
        else:
            model.train_nfold(x_train, y_train_class_rank)
        model.eval(x_test, y_test_class_rank)

    # saving the model
    # model.save()


# classify a list of texts
def classify(texts, output_format, embeddings_name=None, architecture="gru", transformer=None):
    # load model
    model = Classifier("software_context_" + architecture)
    model.load()
    start_time = time.time()
    result = model.predict(texts, output_format)
    runtime = round(time.time() - start_time, 3)
    if output_format == "json":
        result["runtime"] = runtime
    else:
        print("runtime: %s seconds " % (runtime))
    return result


def report_training_contexts(y):
    nb_used = 0
    nb_creation = 0
    nb_shared = 0
    for the_class in y:
        if the_class[0] == 1.0:
            nb_used += 1
        if the_class[1] == 1.0:
            nb_creation += 1
        if the_class[2] == 1.0:
            nb_shared += 1

    print("\ntotal context training cases:", len(y))

    print("\t- used contexts:", nb_used)
    print("\t  not used contexts:", str(len(y) - nb_used))

    print("\t- creation contexts:", nb_creation)
    print("\t  not creation contexts:", str(len(y) - nb_creation))

    print("\t- shared contexts:", nb_shared)
    print("\t  not shared contexts:", str(len(y) - nb_shared))


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Classify the context of a mentioned software using the DeLFT library")

    word_embeddings_examples = ["glove-840B", "fasttext-crawl", "word2vec"]
    pretrained_transformers_examples = ["bert-base-cased", "bert-large-cased", "allenai/scibert_scivocab_cased"]

    parser.add_argument("action")
    parser.add_argument("--fold-count", type=int, default=1)
    parser.add_argument(
        "--architecture", default="gru", help="type of model architecture to be used, one of " + str(architectures)
    )
    parser.add_argument(
        "--embedding",
        default=None,
        help="The desired pre-trained word embeddings using their descriptions in the file. "
        + "For local loading, use delft/resources-registry.json. "
        + "Be sure to use here the same name as in the registry, e.g. "
        + str(word_embeddings_examples)
        + " and that the path in the registry to the embedding file is correct on your system.",
    )
    parser.add_argument(
        "--transformer",
        default=None,
        help="The desired pre-trained transformer to be used in the selected architecture. "
        + "For local loading use, delft/resources-registry.json, and be sure to use here the same name as in the registry, e.g. "
        + str(pretrained_transformers_examples)
        + " and that the path in the registry to the model path is correct on your system. "
        + "HuggingFace transformers hub will be used otherwise to fetch the model, see https://huggingface.co/models "
        + "for model names",
    )

    args = parser.parse_args()

    if args.action not in ("train", "train_eval", "classify", "train_binary", "train_eval_binary"):
        print("action not specified, must be one of [train,train_binary,train_eval,train_eval_binary,classify]")

    embeddings_name = args.embedding
    transformer = args.transformer

    architecture = args.architecture
    if architecture not in architectures:
        print("unknown model architecture, must be one of " + str(architectures))

    if transformer is None and embeddings_name is None:
        # default word embeddings
        embeddings_name = "glove-840B"

    if args.action == "train":
        if args.fold_count < 1:
            raise ValueError("fold-count should be equal or more than 1")

        train(embeddings_name, args.fold_count, architecture=architecture, transformer=transformer)

    if args.action == "train_binary":
        if args.fold_count < 1:
            raise ValueError("fold-count should be equal or more than 1")

        train_binary(embeddings_name, args.fold_count, architecture=architecture, transformer=transformer)

    if args.action == "train_eval":
        if args.fold_count < 1:
            raise ValueError("fold-count should be equal or more than 1")

        y_test = train_and_eval(embeddings_name, args.fold_count, architecture=architecture, transformer=transformer)

    if args.action == "train_eval_binary":
        if args.fold_count < 1:
            raise ValueError("fold-count should be equal or more than 1")

        y_test = train_and_eval_binary(
            embeddings_name, args.fold_count, architecture=architecture, transformer=transformer
        )

    if args.action == "classify":
        someTexts = [
            "Radiographic errors were recorded on individual tick sheets and the information was captured in an Excel spreadsheet (Microsoft, Redmond, WA).",
            "This ground-up approach sits in stark contrast to the top-down one used to introduce patient access to the NHS Summary Care Record via HealthSpace, which has so far met with limited success.",
            "The authors of the GeneWiki project have developed the WikiTrust resource (3), which works via a Firefox plug-in, to mark up Wikipedia articles according to the Wikipedian's reputation.",
        ]
        result = classify(
            someTexts, "json", architecture=architecture, embeddings_name=embeddings_name, transformer=transformer
        )
        print(json.dumps(result, sort_keys=False, indent=4, ensure_ascii=False))



import json

# Model parameters


class ModelConfig(object):
    DEFAULT_FEATURES_VOCABULARY_SIZE = 12
    DEFAULT_FEATURES_EMBEDDING_SIZE = 4

    def __init__(
        self,
        model_name="",
        architecture="BidLSTM_CRF",
        embeddings_name="glove-840B",
        word_embedding_size=300,
        char_emb_size=25,
        char_lstm_units=25,
        max_char_length=30,
        word_lstm_units=100,
        max_sequence_length=300,
        dropout=0.5,
        recurrent_dropout=0.3,
        use_crf=False,
        use_chain_crf=False,
        fold_number=1,
        batch_size=64,
        features_vocabulary_size=DEFAULT_FEATURES_VOCABULARY_SIZE,
        features_indices=None,
        features_embedding_size=DEFAULT_FEATURES_EMBEDDING_SIZE,
        features_lstm_units=DEFAULT_FEATURES_EMBEDDING_SIZE,
        transformer_name=None,
    ):

        self.model_name = model_name
        self.architecture = architecture
        self.embeddings_name = embeddings_name

        self.char_vocab_size = None
        self.case_vocab_size = None

        self.char_embedding_size = char_emb_size
        self.num_char_lstm_units = char_lstm_units
        self.max_char_length = max_char_length

        # Features
        self.features_vocabulary_size = features_vocabulary_size  # maximum number of unique values per feature
        self.features_indices = features_indices
        self.features_embedding_size = features_embedding_size
        self.features_lstm_units = features_lstm_units

        self.max_sequence_length = max_sequence_length
        self.word_embedding_size = word_embedding_size
        self.num_word_lstm_units = word_lstm_units

        self.case_embedding_size = 5
        self.dropout = dropout
        self.recurrent_dropout = recurrent_dropout

        self.use_crf = use_crf
        self.use_chain_crf = use_chain_crf
        self.fold_number = fold_number
        self.batch_size = batch_size  # this is the batch size for prediction

        self.transformer_name = transformer_name

    def save(self, file):
        with open(file, "w") as f:
            json.dump(vars(self), f, sort_keys=False, indent=4)

    @classmethod
    def load(cls, file):
        with open(file) as f:
            variables = json.load(f)
            self = cls()
            for key, val in variables.items():
                setattr(self, key, val)
        return self


# Training parameters
class TrainingConfig(object):
    def __init__(
        self,
        learning_rate,
        batch_size=20,
        optimizer="adam",
        lr_decay=0.9,
        clip_gradients=5.0,
        max_epoch=50,
        early_stop=True,
        patience=5,
        max_checkpoints_to_keep=0,
        multiprocessing=True,
        num_workers=1,
    ):

        self.batch_size = batch_size  # this is the batch size for training
        self.optimizer = optimizer
        self.learning_rate = learning_rate
        self.lr_decay = lr_decay
        self.clip_gradients = clip_gradients
        self.max_epoch = max_epoch
        self.early_stop = early_stop
        self.patience = patience
        self.max_checkpoints_to_keep = max_checkpoints_to_keep
        self.multiprocessing = multiprocessing
        self.num_workers = num_workers



# some convenient methods for all models
import numpy as np
import pandas as pd
import regex as re

# seed is fixed for reproducibility
from numpy.random import seed

seed(7)
import argparse
import os.path
import shutil
from urllib.parse import urlparse

import requests
import truecase
from tf_keras.preprocessing import text
from tqdm import tqdm


def truncate_batch_values(batch_values: list, max_sequence_length: int) -> list:
    return [row[:max_sequence_length] for row in batch_values]


# read list of words (one per line), e.g. stopwords, badwords
def read_words(words_file):
    return [line.replace("\n", "").lower() for line in open(words_file, "r")]


# preprocessing used for twitter-trained glove embeddings
def glove_preprocess(text):
    """
    adapted from https://nlp.stanford.edu/projects/glove/preprocess-twitter.rb

    """
    # Different regex parts for smiley faces
    eyes = "[8:=;]"
    nose = r"['`\-]?"
    text = re.sub("https?:* ", "<URL>", text)
    text = re.sub("www.* ", "<URL>", text)
    text = re.sub(r"\[\[User(.*)\|", "<USER>", text)
    text = re.sub("<3", "<HEART>", text)
    text = re.sub(r"[-+]?[.\d]*[\d]+[:,.\d]*", "<NUMBER>", text)
    text = re.sub(eyes + nose + "[Dd)]", "<SMILE>", text)
    text = re.sub("[(d]" + nose + eyes, "<SMILE>", text)
    text = re.sub(eyes + nose + "p", "<LOLFACE>", text)
    text = re.sub(eyes + nose + r"\(", "<SADFACE>", text)
    text = re.sub(r"\)" + nose + eyes, "<SADFACE>", text)
    text = re.sub(eyes + nose + "[/|l*]", "<NEUTRALFACE>", text)
    text = re.sub("/", " / ", text)
    text = re.sub(r"[-+]?[.\d]*[\d]+[:,.\d]*", "<NUMBER>", text)
    text = re.sub("([!]){2,}", "! <REPEAT>", text)
    text = re.sub("([?]){2,}", "? <REPEAT>", text)
    text = re.sub("([.]){2,}", ". <REPEAT>", text)
    pattern = re.compile(r"(.)\1{2,}")
    text = pattern.sub(r"\1" + " <ELONG>", text)

    return text


# split provided sequence data in two sets given the given ratio between 0 and 1
# for instance ratio at 0.8 will split 80% of the sentence in the first set and 20%
# of the remaining sentence in the second one
#
def split_data_and_labels(x, y, ratio):
    if len(x) != len(y):
        print("error: size of x and y set must be equal, ", len(x), "=/=", len(y))
        return
    x1 = []
    x2 = []
    y1 = []
    y2 = []
    for i in range(len(x)):
        if np.random.random_sample() < ratio:
            x1.append(x[i])
            y1.append(y[i])
        else:
            x2.append(x[i])
            y2.append(y[i])
    return np.asarray(x1), np.asarray(y1), np.asarray(x2), np.asarray(y2)


url_regex = re.compile(
    r"https?:\/\/[a-zA-Z0-9_\-\.]+(?:com|org|fr|de|uk|se|net|edu|gov|int|mil|biz|info|br|ca|cn|in|jp|ru|au|us|ch|it|nl|no|es|pl|ir|cz|kr|co|gr|za|tw|hu|vn|be|mx|at|tr|dk|me|ar|fi|nz)\/?\b"
)


# produce some statistics
def stats(x_train=None, y_train=None, x_valid=None, y_valid=None, x_eval=None, y_eval=None):
    charset = []
    nb_total_sequences = 0
    nb_total_tokens = 0
    if x_train is not None:
        print(len(x_train), "train sequences")
        nb_total_sequences += len(x_train)
        nb_tokens = 0
        for sentence in x_train:
            nb_tokens += len(sentence)
            for token in sentence:
                for character in token:
                    if character not in charset:
                        charset.append(character)
        print("\t", "nb. tokens", nb_tokens)
        nb_total_tokens += nb_tokens
    if y_train is not None:
        nb_entities = 0
        for labels in y_train:
            for label in labels:
                if label != "O":
                    nb_entities += 1
        print("\t", "with nb. entities", nb_entities)
    if x_valid is not None:
        print(len(x_valid), "validation sequences")
        nb_total_sequences += len(x_valid)
        nb_tokens = 0
        for sentence in x_valid:
            nb_tokens += len(sentence)
            for token in sentence:
                for character in token:
                    if character not in charset:
                        charset.append(character)
        print("\t", "nb. tokens", nb_tokens)
        nb_total_tokens += nb_tokens
    if y_valid is not None:
        nb_entities = 0
        for labels in y_valid:
            for label in labels:
                if label != "O":
                    nb_entities += 1
        print("\t", "with nb. entities", nb_entities)
    if x_eval is not None:
        print(len(x_eval), "evaluation sequences")
        nb_total_sequences += len(x_eval)
        nb_tokens = 0
        for sentence in x_eval:
            nb_tokens += len(sentence)
            for token in sentence:
                for character in token:
                    if character not in charset:
                        charset.append(character)
        print("\t", "nb. tokens", nb_tokens)
        nb_total_tokens += nb_tokens
    if y_eval is not None:
        nb_entities = 0
        for labels in y_eval:
            for label in labels:
                if label != "O":
                    nb_entities += 1
        print("\t", "with nb. entities", nb_entities)

    print("\n")
    print(nb_total_sequences, "total sequences")
    print(nb_total_tokens, "total tokens\n")

    print("total distinct characters:", len(charset), "\n")
    # print(charset)


# generate the list of out of vocabulary words present in the Toxic dataset
# with respect to 3 embeddings: fastText, Gloves and word2vec
def generateOOVEmbeddings():
    # read the (DL cleaned) dataset and build the vocabulary
    print("loading dataframes...")
    train_df = pd.read_csv("../data/training/train2.cleaned.dl.csv")
    test_df = pd.read_csv("../data/eval/test2.cleaned.dl.csv")

    # ps: forget memory and runtime, it's python here :D
    list_sentences_train = train_df["comment_text"].values
    list_sentences_test = test_df["comment_text"].values
    list_sentences_all = np.concatenate([list_sentences_train, list_sentences_test])

    tokenizer = text.Tokenizer(num_words=400000)
    tokenizer.fit_on_texts(list(list_sentences_all))
    print("word_index size:", len(tokenizer.word_index), "words")
    word_index = tokenizer.word_index

    # load fastText - only the words
    print("loading fastText embeddings...")
    voc = set()
    f = open("/mnt/data/wikipedia/embeddings/crawl-300d-2M.vec")
    begin = True
    for line in f:
        if begin:
            begin = False
        else:
            values = line.split()
            word = " ".join(values[:-300])
            voc.add(word)
    f.close()
    print("fastText embeddings:", len(voc), "words")

    oov = []
    for tokenStr in word_index:
        if tokenStr not in voc:
            oov.append(tokenStr)

    print("fastText embeddings:", len(oov), "out-of-vocabulary")

    with open("../data/training/oov-fastText.txt", "w") as oovFile:
        for w in oov:
            oovFile.write(w)
            oovFile.write("\n")
    oovFile.close()

    # load gloves - only the words
    print("loading gloves embeddings...")
    voc = set()
    f = open("/mnt/data/wikipedia/embeddings/glove.840B.300d.txt")
    for line in f:
        values = line.split()
        word = " ".join(values[:-300])
        voc.add(word)
    f.close()
    print("gloves embeddings:", len(voc), "words")

    oov = []
    for tokenStr in word_index:
        if tokenStr not in voc:
            oov.append(tokenStr)

    print("gloves embeddings:", len(oov), "out-of-vocabulary")

    with open("../data/training/oov-gloves.txt", "w") as oovFile:
        for w in oov:
            oovFile.write(w)
            oovFile.write("\n")
    oovFile.close()

    # load word2vec - only the words
    print("loading word2vec embeddings...")
    voc = set()
    f = open("/mnt/data/wikipedia/embeddings/GoogleNews-vectors-negative300.vec")
    begin = True
    for line in f:
        if begin:
            begin = False
        else:
            values = line.split()
            word = " ".join(values[:-300])
            voc.add(word)
    f.close()
    print("word2vec embeddings:", len(voc), "words")

    oov = []
    for tokenStr in word_index:
        if tokenStr not in voc:
            oov.append(tokenStr)

    print("word2vec embeddings:", len(oov), "out-of-vocabulary")

    with open("../data/training/oov-w2v.txt", "w") as oovFile:
        for w in oov:
            oovFile.write(w)
            oovFile.write("\n")
    oovFile.close()

    # load numberbatch - only the words
    print("loading numberbatch embeddings...")
    voc = set()
    f = open("/mnt/data/wikipedia/embeddings/numberbatch-en-17.06.txt")
    begin = True
    for line in f:
        if begin:
            begin = False
        else:
            values = line.split()
            word = " ".join(values[:-300])
            voc.add(word)
    f.close()
    print("numberbatch embeddings:", len(voc), "words")

    oov = []
    for tokenStr in word_index:
        if tokenStr not in voc:
            oov.append(tokenStr)

    print("numberbatch embeddings:", len(oov), "out-of-vocabulary")

    with open("../data/training/oov-numberbatch.txt", "w") as oovFile:
        for w in oov:
            oovFile.write(w)
            oovFile.write("\n")
    oovFile.close()


def ontonotes_conll2012_names(pathin, pathout):
    # generate the list of files having a .name extension in the complete ontonotes corpus
    fileout = open(os.path.join(pathout, "names.list"), "w+")

    for subdir, dirs, files in os.walk(pathin):
        for file in files:
            if file.endswith(".name"):
                ind = subdir.find("data/english/")
                if ind == -1:
                    print("path to ontonotes files appears invalid")
                subsubdir = subdir[ind:]
                fileout.write(os.path.join(subsubdir, file.replace(".name", "")))
                fileout.write("\n")
    fileout.close()


def convert_conll2012_to_iob2(pathin, pathout):
    """
    This method will post-process the assembled Ontonotes CoNLL-2012 data for NER.
    It will take an input like:
      bc/cctv/00/cctv_0001   0    5         Japanese    JJ             *           -    -      -   Speaker#1    (NORP)           *        *            *        *     -
    and transform it into a simple and readable:
      Japanese  B-NORP
    taking into account the sequence markers and an expected IOB2 scheme.
    """
    if pathin == pathout:
        print("input and ouput path must be different:", pathin, pathout)
        return

    names_doc_ids = []
    with open(os.path.join("data", "sequenceLabelling", "CoNLL-2012-NER", "names.list"), "r") as f:
        for line in f:
            line = line.rstrip()
            if len(line) == 0:
                continue
            names_doc_ids.append(line)
    print("number of documents with name notation:", len(names_doc_ids))

    nb_files = 0
    # first pass to get number of files - test files for CoNLL-2012 are under conll-2012-test/, not test/
    # we ignore files not having .names extension in the original ontonotes realease
    for subdir, dirs, files in os.walk(pathin):
        for file in files:
            if (
                "/english/" in subdir
                and (file.endswith("gold_conll"))
                and "/pt/" not in subdir
                and "/test/" not in subdir
            ):
                ind = subdir.find("data/english/")
                if ind == -1:
                    print("path to ontonotes files appears invalid")
                subsubdir = os.path.join(subdir[ind:], file.replace(".gold_conll", ""))
                if subsubdir in names_doc_ids:
                    nb_files += 1
    nb_total_files = nb_files
    print(nb_total_files, "total files to convert")

    # load restricted set of ids for the CoNLL-2012 dataset
    train_doc_ids = []
    dev_doc_ids = []
    test_doc_ids = []

    with open(
        os.path.join("data", "sequenceLabelling", "CoNLL-2012-NER", "english-ontonotes-5.0-train-document-ids.txt"), "r"
    ) as f:
        for line in f:
            line = line.rstrip()
            if len(line) == 0:
                continue
            train_doc_ids.append(line)
    print("number of train documents:", len(train_doc_ids))

    with open(
        os.path.join(
            "data", "sequenceLabelling", "CoNLL-2012-NER", "english-ontonotes-5.0-development-document-ids.txt"
        ),
        "r",
    ) as f:
        for line in f:
            line = line.rstrip()
            if len(line) == 0:
                continue
            dev_doc_ids.append(line)
    print("number of development documents:", len(dev_doc_ids))

    with open(
        os.path.join(
            "data", "sequenceLabelling", "CoNLL-2012-NER", "english-ontonotes-5.0-conll-2012-test-document-ids.txt"
        ),
        "r",
    ) as f:
        for line in f:
            line = line.rstrip()
            if len(line) == 0:
                continue
            test_doc_ids.append(line)
    print("number of test documents:", len(test_doc_ids))

    train_out = open(os.path.join(pathout, "eng.train"), "w+", encoding="UTF-8")
    dev_out = open(os.path.join(pathout, "eng.dev"), "w+", encoding="UTF-8")
    test_out = open(os.path.join(pathout, "eng.test"), "w+", encoding="UTF-8")

    nb_files = 0
    pbar = tqdm(total=nb_total_files)
    for subdir, dirs, files in os.walk(pathin):
        for file in files:
            # pt subdirectory corresponds to the old and new testaments, it does not contain NER annotation, so it is traditionally ignored
            # if '/english/' in subdir and (file.endswith('gold_conll') or ('/test/' in subdir and file.endswith('gold_parse_conll'))) and not '/pt/' in subdir:
            if (
                "/english/" in subdir
                and (file.endswith("gold_conll"))
                and "/pt/" not in subdir
                and "/test/" not in subdir
            ):
                ind = subdir.find("data/english/")
                if ind == -1:
                    print("path to ontonotes files appears invalid")
                subsubdir = os.path.join(subdir[ind:], file.replace(".gold_conll", ""))

                if subsubdir not in names_doc_ids:
                    continue

                pbar.update(1)

                f2 = None
                if "/train/" in subdir and subsubdir in train_doc_ids:
                    f2 = train_out
                elif "/development/" in subdir and subsubdir in dev_doc_ids:
                    f2 = dev_out
                elif "/conll-2012-test/" in subdir and subsubdir in test_doc_ids:
                    f2 = test_out

                if f2 is None:
                    continue

                with open(os.path.join(subdir, file), "r", encoding="UTF-8") as f1:
                    previous_tag = None
                    for line in f1:
                        line_ = line.rstrip()
                        line_ = " ".join(line_.split())
                        if len(line_) == 0:
                            f2.write("\n")
                            previous_tag = None
                        elif line_.startswith("#begin document"):
                            f2.write(line_ + "\n\n")
                            previous_tag = None
                        elif line_.startswith("#end document"):
                            # f2.write("\n")
                            previous_tag = None
                        else:
                            pieces = line_.split(" ")
                            if len(pieces) < 11:
                                print(
                                    os.path.join(subdir, file),
                                    "-> unexpected number of fiels for line (",
                                    len(pieces),
                                    "):",
                                    line_,
                                )
                                previous_tag = None
                            word = pieces[3]
                            # some punctuation are prefixed by / (e.g. /. or /? for dialogue turn apparently)
                            if word.startswith("/") and len(word) > 1:
                                word = word[1:]
                            # in dialogue texts, interjections are maked with a prefix %, e.g. %uh, %eh, we remove this prefix
                            if word.startswith("%") and len(word) > 1:
                                word = word[1:]
                            # there are '='' prefixes to some words, although I don't know what it is used for, we remove it
                            if word.startswith("=") and len(word) > 1:
                                word = word[1:]
                            # we have some markers like -LRB- left bracket, -RRB- right bracket
                            if word == "-LRB-":
                                word = "("
                            if word == "-RRB-":
                                word = ")"
                            # some tokens are identifier in the form 165.00_177.54_B:, 114.86_118.28_A:, and so on, always _A or _B as suffix
                            # it's very unclear why it is in the plain text but clearly noise
                            # regex_str = "\d\d\d\.\d\d_\d\d\d\.\d\d_(A|B)"

                            tag = pieces[10]
                            if tag.startswith("("):
                                if tag.endswith(")"):
                                    tag = tag[1:-1]
                                    previous_tag = None
                                else:
                                    tag = tag[1:-1]
                                    previous_tag = tag
                                f2.write(word + "\tB-" + tag + "\n")
                            elif tag == "*" and previous_tag is not None:
                                f2.write(word + "\tI-" + previous_tag + "\n")
                            elif tag == "*)":
                                f2.write(word + "\tI-" + previous_tag + "\n")
                                previous_tag = None
                            else:
                                f2.write(word + "\tO\n")
                                previous_tag = None
    pbar.close()

    train_out.close()
    dev_out.close()
    test_out.close()


def convert_conll2003_to_iob2(filein, fileout):
    """
    This method will post-process the assembled CoNLL-2003 data for NER.
    It will take an input like:

    and transform it into a simple and readable:
      Japanese  B-NORP
    taking into account the sequence markers and an expected IOB2 scheme.
    """
    with open(filein, "r") as f1:
        with open(fileout, "w") as f2:
            previous_tag = "O"
            for line in f1:
                line_ = line.rstrip()
                if len(line_) == 0 or line_.startswith("-DOCSTART-"):
                    f2.write(line_ + "\n")
                    previous_tag = "O"
                else:
                    word, pos, phrase, tag = line_.split(" ")
                    if tag == "O" or tag.startswith("B-"):
                        f2.write(word + "\t" + tag + "\n")
                    else:
                        tag[2:]
                        if previous_tag.endswith(tag[2:]):
                            f2.write(word + "\t" + tag + "\n")
                        else:
                            f2.write(word + "\tB-" + tag[2:] + "\n")
                    previous_tag = tag


def merge_folders(root_src_dir, root_dst_dir):
    """
    Recursively merge two folders including subfolders.
    This method is motivated by the limitation of shutil.copytree() which supposes that the
    destination directory must not exist.
    """
    for src_dir, dirs, files in os.walk(root_src_dir):
        dst_dir = src_dir.replace(root_src_dir, root_dst_dir, 1)
        if not os.path.exists(dst_dir):
            os.makedirs(dst_dir)
        for file_ in files:
            src_file = os.path.join(src_dir, file_)
            dst_file = os.path.join(dst_dir, file_)
            if os.path.exists(dst_file):
                os.remove(dst_file)
            shutil.copy(src_file, dst_dir)


def truecase_sentence(tokens):
    """
    from https://github.com/ghaddarAbs
    for experimenting with CoNLL-2003 casing
    """
    word_lst = [(w, idx) for idx, w in enumerate(tokens) if all(c.isalpha() for c in w)]
    lst = [w for w, _ in word_lst if re.match(r"\b[A-Z\.\-]+\b", w)]

    if len(lst) and len(lst) == len(word_lst):
        parts = truecase.get_true_case(" ".join(lst)).split()

        # the trucaser have its own tokenization ...
        # skip if the number of word doesn't match
        if len(parts) != len(word_lst):
            return tokens

        for (w, idx), nw in zip(word_lst, parts):
            tokens[idx] = nw
    return tokens


def download_file(url, path, filename=None):
    """
    Download with Python requests which handle well compression
    """
    # check path
    if path is None or not os.path.isdir(path):
        print("Invalid destination directory:", path)
    HEADERS = {"""User-Agent""": """Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:81.0) Gecko/20100101 Firefox/81.0"""}
    result = "fail"
    print("downloading", url)

    if filename is None:
        # we use the actual server file name
        a = urlparse(url)
        filename = os.path.basename(a.path)
    destination = os.path.join(path, filename)
    try:
        resp = requests.get(url, stream=True, allow_redirects=True, headers=HEADERS)
        total_length = resp.headers.get("content-length")

        if total_length is None and resp.status_code == 200:
            # no content length header available, can't have a progress bar :(
            with open(destination, "wb") as f_out:
                f_out.write(resp.content)
        elif resp.status_code == 200:
            total = int(total_length)
            with (
                open(destination, "wb") as f_out,
                tqdm(
                    desc=destination,
                    total=total,
                    unit="iB",
                    unit_scale=True,
                    unit_divisor=1024,
                ) as bar,
            ):
                for data in resp.iter_content(chunk_size=1024):
                    size = f_out.write(data)
                    bar.update(size)
            result = "success"
    except Exception as e:
        print("Download failed for {0} with requests".format(url), "\nError:", e)
    if result == "success":
        return destination
    else:
        return None


def len_until_first_pad(tokens, pad):
    i = len(tokens) - 1
    while i >= 0:
        if tokens[i] != pad:
            return i + 1
        i -= 1
    return 0


def len_until_first_pad_old(tokens, pad):
    for i in range(len(tokens)):
        if tokens[i] == pad:
            return i
    return len(tokens)


def longest_row(array):
    the_longest_row = max(len(row) for row in array)
    return the_longest_row


if __name__ == "__main__":
    # usage example - for CoNLL-2003, indicate the eng.* file to be converted:
    # > python3 utilities/Utilities.py --dataset-type conll2003 --data-path /home/lopez/resources/CoNLL-2003/eng.train --output-path /home/lopez/resources/CoNLL-2003/iob2/eng.train
    # for CoNLL-2012, indicate the root directory of the ontonotes data (in CoNLL-2012 format) to be converted:
    # > python3 utilities/Utilities.py --dataset-type conll2012 --data-path /home/lopez/resources/ontonotes/conll-2012/ --output-path /home/lopez/resources/ontonotes/conll-2012/iob2/

    # get the argument
    parser = argparse.ArgumentParser(description="Named Entity Recognizer dataset converter to OIB2 tagging scheme")

    # parser.add_argument("action")
    parser.add_argument(
        "--dataset-type",
        default="conll2003",
        help="dataset to be used for training the model, one of ['conll2003','conll2012']",
    )
    parser.add_argument("--data-path", default=None, help="path to the corpus of documents to process")
    parser.add_argument("--output-path", default=None, help="path to write the converted dataset")

    args = parser.parse_args()

    # action = args.action
    dataset_type = args.dataset_type
    data_path = args.data_path
    output_path = args.output_path

    if dataset_type == "conll2003":
        convert_conll2003_to_iob2(data_path, output_path)
    elif dataset_type == "conll2012":
        convert_conll2012_to_iob2(data_path, output_path)
    elif dataset_type == "ontonotes":
        ontonotes_conll2012_names(data_path, output_path)


def t_or_f(arg):
    ua = str(arg).upper()
    if "TRUE".startswith(ua):
        return True
    elif "FALSE".startswith(ua):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected. Omit this option to use default values.")



import argparse
import json
import time

from delft.textClassification import Classifier
from delft.textClassification.models import architectures
from delft.textClassification.reader import load_software_use_corpus_json
from delft.utilities.Utilities import split_data_and_labels

"""
    This binary classifier is used in combination with a software mention recognition model, for characterizing
    the nature of the citation of software in scientific and technical literature.
    This classifier predicts if the software introduced by a software mention in a sentence is used
    or not by the described work.

    For the software mention recognizer, see https://github.com/ourresearch/software-mentions
    and grobidTagger.py in the present project DeLFT.

    Best architecture/model is fine-tuned SciBERT.
"""

list_classes = ["not_used", "used"]

class_weights = {0: 1.0, 1: 4.0}


def configure(architecture):
    batch_size = 256
    maxlen = 300
    patience = 5
    early_stop = True
    max_epoch = 60

    # default bert model parameters
    if architecture == "bert":
        batch_size = 32
        early_stop = False
        max_epoch = 3
        maxlen = 100

    return batch_size, maxlen, patience, early_stop, max_epoch


def train(embeddings_name, fold_count, architecture="gru", transformer=None):
    print("loading binary software use dataset...")
    xtr, y = load_software_use_corpus_json("data/textClassification/software/software-use.json.gz")

    model_name = "software_use_" + architecture
    class_weights = None

    batch_size, maxlen, patience, early_stop, max_epoch = configure(architecture)

    model = Classifier(
        model_name,
        architecture=architecture,
        list_classes=list_classes,
        max_epoch=max_epoch,
        fold_number=fold_count,
        patience=patience,
        use_roc_auc=True,
        embeddings_name=embeddings_name,
        batch_size=batch_size,
        maxlen=maxlen,
        early_stop=early_stop,
        class_weights=class_weights,
        transformer_name=transformer,
    )

    if fold_count == 1:
        model.train(xtr, y)
    else:
        model.train_nfold(xtr, y)
    # saving the model
    model.save()


def train_and_eval(embeddings_name, fold_count, architecture="gru", transformer=None):
    print("loading binary software use dataset...")
    xtr, y = load_software_use_corpus_json("data/textClassification/software/software-use.json.gz")

    nb_used = 0
    for the_class in y:
        if the_class[1] == 1.0:
            nb_used += 1
    nb_unused = len(y) - nb_used
    print("\ttotal:", len(y))
    print("\tused:", nb_used)
    print("\tnot used:", nb_unused)

    model_name = "software_use_" + architecture
    class_weights = None

    # segment train and eval sets
    x_train, y_train, x_test, y_test = split_data_and_labels(xtr, y, 0.9)

    batch_size, maxlen, patience, early_stop, max_epoch = configure(architecture)

    print(list_classes)

    model = Classifier(
        model_name,
        architecture=architecture,
        list_classes=list_classes,
        max_epoch=max_epoch,
        fold_number=fold_count,
        patience=patience,
        use_roc_auc=True,
        embeddings_name=embeddings_name,
        batch_size=batch_size,
        maxlen=maxlen,
        early_stop=early_stop,
        class_weights=class_weights,
        transformer_name=transformer,
    )

    if fold_count == 1:
        model.train(x_train, y_train)
    else:
        model.train_nfold(x_train, y_train)
    model.eval(x_test, y_test)

    # saving the model
    model.save()


# classify a list of texts
def classify(texts, output_format, embeddings_name=None, architecture="gru", transformer=None):
    # load model
    model = Classifier(
        "software_use_" + architecture,
        architecture=architecture,
        list_classes=list_classes,
        embeddings_name=embeddings_name,
        transformer_name=transformer,
    )
    model.load()
    start_time = time.time()
    result = model.predict(texts, output_format)
    runtime = round(time.time() - start_time, 3)
    if output_format == "json":
        result["runtime"] = runtime
    else:
        print("runtime: %s seconds " % (runtime))
    return result


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Classify whether a mentioned software is used or not, using the DeLFT library"
    )

    word_embeddings_examples = ["glove-840B", "fasttext-crawl", "word2vec"]
    pretrained_transformers_examples = ["bert-base-cased", "bert-large-cased", "allenai/scibert_scivocab_cased"]

    parser.add_argument("action")
    parser.add_argument("--fold-count", type=int, default=1)
    parser.add_argument(
        "--architecture", default="gru", help="type of model architecture to be used, one of " + str(architectures)
    )
    parser.add_argument(
        "--embedding",
        default=None,
        help="The desired pre-trained word embeddings using their descriptions in the file. "
        + "For local loading, use delft/resources-registry.json. "
        + "Be sure to use here the same name as in the registry, e.g. "
        + str(word_embeddings_examples)
        + " and that the path in the registry to the embedding file is correct on your system.",
    )
    parser.add_argument(
        "--transformer",
        default=None,
        help="The desired pre-trained transformer to be used in the selected architecture. "
        + "For local loading use, delft/resources-registry.json, and be sure to use here the same name as in the registry, e.g. "
        + str(pretrained_transformers_examples)
        + " and that the path in the registry to the model path is correct on your system. "
        + "HuggingFace transformers hub will be used otherwise to fetch the model, see https://huggingface.co/models "
        + "for model names",
    )

    args = parser.parse_args()

    if args.action not in ("train", "train_eval", "classify"):
        print("action not specified, must be one of [train,train_eval,classify]")

    embeddings_name = args.embedding
    transformer = args.transformer

    architecture = args.architecture
    if architecture not in architectures:
        print("unknown model architecture, must be one of " + str(architectures))

    if transformer is None and embeddings_name is None:
        # default word embeddings
        embeddings_name = "glove-840B"

    if args.action == "train":
        if args.fold_count < 1:
            raise ValueError("fold-count should be equal or more than 1")

        train(embeddings_name, args.fold_count, architecture=architecture, transformer=transformer)

    if args.action == "train_eval":
        if args.fold_count < 1:
            raise ValueError("fold-count should be equal or more than 1")

        y_test = train_and_eval(embeddings_name, args.fold_count, architecture=architecture, transformer=transformer)

    if args.action == "classify":
        someTexts = [
            "Radiographic errors were recorded on individual tick sheets and the information was captured in an Excel spreadsheet (Microsoft, Redmond, WA).",
            "This ground-up approach sits in stark contrast to the top-down one used to introduce patient access to the NHS Summary Care Record via HealthSpace, which has so far met with limited success.",
            "The authors of the GeneWiki project have developed the WikiTrust resource (3), which works via a Firefox plug-in, to mark up Wikipedia articles according to the Wikipedian's reputation.",
        ]
        result = classify(
            someTexts, "json", architecture=architecture, embeddings_name=embeddings_name, transformer=transformer
        )
        print(json.dumps(result, sort_keys=False, indent=4, ensure_ascii=False))



import argparse
import json
import time

import numpy as np

from delft.textClassification import Classifier
from delft.textClassification.models import architectures
from delft.textClassification.reader import load_texts_and_classes_pandas_no_id
from delft.textClassification.reader import vectorize as vectorizer
from delft.utilities.numpy import shuffle_triple_with_view
from delft.utilities.Utilities import split_data_and_labels

"""
    Two multiclass classifiers to be used in combination with Grobid to classify a license/copyrights section
    extracted from a scientific article into two dimensions:
    - copyright owner: publisher, authors or undecidable (changed from NA to avoid issues with pandas)
    - license associated to the article file: explicit copyrights (no restriction), creative commons licenses
      (CC-0, CC-BY, CC-BY-NC, etc.), other, or undecidable (changed from NA to avoid issues with pandas)

    Note: when the license is undecidable, this means normal copyrights when a copyright owner exists.
"""

list_classes_copyright = ["publisher", "authors", "undecided"]
class_weights_copyright = {0: 1.0, 1: 1.0, 2: 1.0}
list_classes_licenses = [
    "CC-0",
    "CC-BY",
    "CC-BY-NC",
    "CC-BY-NC-ND",
    "CC-BY-SA",
    "CC-BY-NC-SA",
    "CC-BY-ND",
    "copyright",
    "other",
    "undecided",
]
class_weights_licenses = {0: 1.0, 1: 1.0, 2: 1.0, 3: 1.0, 4: 1.0, 5: 1.0, 6: 1.0, 7: 1.0, 8: 1.0, 9: 1.0}


def configure(architecture):
    batch_size = 256
    maxlen = 300
    patience = 5
    early_stop = True
    max_epoch = 60

    # default bert model parameters
    if architecture == "bert":
        batch_size = 16
        early_stop = False
        max_epoch = 6
        maxlen = 200

    return batch_size, maxlen, patience, early_stop, max_epoch


def train(embeddings_name, fold_count, architecture="gru", transformer=None):
    print("loading multiclass copyright/license dataset...")
    xtr, y_copyrights = _read_data(
        "data/textClassification/licenses/copyrights-licenses-data-validated.csv", data_type="copyrights"
    )

    report_training_copyrights(y_copyrights)

    # copyright ownwer classifier
    model_name = "copyright_" + architecture
    batch_size, maxlen, patience, early_stop, max_epoch = configure(architecture)

    model = Classifier(
        model_name,
        architecture=architecture,
        list_classes=list_classes_copyright,
        max_epoch=max_epoch,
        fold_number=fold_count,
        patience=patience,
        use_roc_auc=True,
        embeddings_name=embeddings_name,
        batch_size=batch_size,
        maxlen=maxlen,
        early_stop=early_stop,
        class_weights=class_weights_copyright,
        transformer_name=transformer,
    )

    if fold_count == 1:
        model.train(xtr, y_copyrights)
    else:
        model.train_nfold(xtr, y_copyrights)
    # saving the model
    model.save()

    # license classifier
    xtr, y_licenses = _read_data(
        "data/textClassification/licenses/copyrights-licenses-data-validated.csv", data_type="licenses"
    )
    model_name = "license_" + architecture
    class_weights_licenses = None
    # batch_size, maxlen, patience, early_stop, max_epoch = configure(architecture)

    model = Classifier(
        model_name,
        architecture=architecture,
        list_classes=list_classes_licenses,
        max_epoch=max_epoch,
        fold_number=fold_count,
        patience=patience,
        use_roc_auc=True,
        embeddings_name=embeddings_name,
        batch_size=batch_size,
        maxlen=maxlen,
        early_stop=early_stop,
        class_weights=class_weights_licenses,
        transformer_name=transformer,
    )

    if fold_count == 1:
        model.train(xtr, y_licenses)
    else:
        model.train_nfold(xtr, y_licenses)
    # saving the model
    model.save()


def train_and_eval(embeddings_name, fold_count, architecture="gru", transformer=None):
    print("loading multiclass copyright/license dataset...")
    xtr, y_copyrights = _read_data(
        "data/textClassification/licenses/copyrights-licenses-data-validated.csv", data_type="copyrights"
    )

    report_training_copyrights(y_copyrights)

    # copyright ownwer classifier
    model_name = "copyright_" + architecture

    # segment train and eval sets
    x_train, y_train, x_test, y_test = split_data_and_labels(xtr, y_copyrights, 0.9)
    batch_size, maxlen, patience, early_stop, max_epoch = configure(architecture)

    print(list_classes_copyright)

    model = Classifier(
        model_name,
        architecture=architecture,
        list_classes=list_classes_copyright,
        max_epoch=max_epoch,
        fold_number=fold_count,
        patience=patience,
        use_roc_auc=True,
        embeddings_name=embeddings_name,
        batch_size=batch_size,
        maxlen=maxlen,
        early_stop=early_stop,
        class_weights=class_weights_copyright,
        transformer_name=transformer,
    )

    if fold_count == 1:
        model.train(x_train, y_train)
    else:
        model.train_nfold(x_train, y_train)
    model.eval(x_test, y_test)

    # saving the model
    model.save()

    # license classifier
    xtr, y_licenses = _read_data(
        "data/textClassification/licenses/copyrights-licenses-data-validated.csv", data_type="licenses"
    )

    model_name = "license_" + architecture

    # segment train and eval sets
    x_train, y_train, x_test, y_test = split_data_and_labels(xtr, y_licenses, 0.9)
    # batch_size, maxlen, patience, early_stop, max_epoch = configure(architecture)
    class_weights_licenses = None

    model = Classifier(
        model_name,
        architecture=architecture,
        list_classes=list_classes_licenses,
        max_epoch=max_epoch,
        fold_number=fold_count,
        patience=patience,
        use_roc_auc=True,
        embeddings_name=embeddings_name,
        batch_size=batch_size,
        maxlen=maxlen,
        early_stop=early_stop,
        class_weights=class_weights_licenses,
        transformer_name=transformer,
    )

    if fold_count == 1:
        model.train(x_train, y_train)
    else:
        model.train_nfold(x_train, y_train)
    model.eval(x_test, y_test)

    # saving the model (not so useful...)
    model.save()


def train_binary(embeddings_name, fold_count, architecture="gru", transformer=None):
    print("loading multiclass copyright/license dataset...")
    xtr, y_copyrights = _read_data(
        "data/textClassification/licenses/copyrights-licenses-data-validated.csv", data_type="copyrights"
    )

    report_training_copyrights(y_copyrights)

    for class_rank in range(len(list_classes_copyright)):
        model_name = "copyright_" + list_classes_copyright[class_rank] + "_" + architecture

        # we could experiment with binary weighting the class maybe
        class_weights = None

        batch_size, maxlen, patience, early_stop, max_epoch = configure(architecture)

        y_train_class_rank = [[1, 0] if y[class_rank] == 1.0 else [0, 1] for y in y_copyrights]
        y_train_class_rank = np.array(y_train_class_rank)

        list_classes_rank = [list_classes_copyright[class_rank], "not_" + list_classes_copyright[class_rank]]

        model = Classifier(
            model_name,
            architecture=architecture,
            list_classes=list_classes_rank,
            max_epoch=max_epoch,
            fold_number=fold_count,
            patience=patience,
            use_roc_auc=True,
            embeddings_name=embeddings_name,
            batch_size=batch_size,
            maxlen=maxlen,
            early_stop=early_stop,
            class_weights=class_weights,
            transformer_name=transformer,
        )

        if fold_count == 1:
            model.train(xtr, y_train_class_rank)
        else:
            model.train_nfold(xtr, y_train_class_rank)
        # saving the model
        model.save()

    xtr, y_licenses = _read_data(
        "data/textClassification/licenses/copyrights-licenses-data-validated.csv", data_type="licenses"
    )
    for class_rank in range(len(list_classes_licenses)):
        model_name = "licenses_" + list_classes_licenses[class_rank] + "_" + architecture

        # we could experiment with binary weighting the class maybe
        class_weights = None

        batch_size, maxlen, patience, early_stop, max_epoch = configure(architecture)

        y_train_class_rank = [[1, 0] if y[class_rank] == 1.0 else [0, 1] for y in y_licenses]
        y_train_class_rank = np.array(y_train_class_rank)

        list_classes_rank = [list_classes_licenses[class_rank], "not_" + list_classes_licenses[class_rank]]

        model = Classifier(
            model_name,
            architecture=architecture,
            list_classes=list_classes_rank,
            max_epoch=max_epoch,
            fold_number=fold_count,
            patience=patience,
            use_roc_auc=True,
            embeddings_name=embeddings_name,
            batch_size=batch_size,
            maxlen=maxlen,
            early_stop=early_stop,
            class_weights=class_weights,
            transformer_name=transformer,
        )

        if fold_count == 1:
            model.train(xtr, y_train_class_rank)
        else:
            model.train_nfold(xtr, y_train_class_rank)
        # saving the model
        model.save()


def train_and_eval_binary(embeddings_name, fold_count, architecture="gru", transformer=None):
    print("loading multiclass copyright/license dataset...")
    xtr, y_copyrights = _read_data(
        "data/textClassification/licenses/copyrights-licenses-data-validated.csv", data_type="copyrights"
    )

    report_training_copyrights(y_copyrights)

    # segment train and eval sets
    x_train, y_train, x_test, y_test = split_data_and_labels(xtr, y_copyrights, 0.9)

    for class_rank in range(len(list_classes_copyright)):
        model_name = "copyright_" + list_classes_copyright[class_rank] + "_" + architecture
        class_weights = None

        batch_size, maxlen, patience, early_stop, max_epoch = configure(architecture)

        y_train_class_rank = [[1, 0] if y[class_rank] == 1.0 else [0, 1] for y in y_train]
        y_test_class_rank = [[1, 0] if y[class_rank] == 1.0 else [0, 1] for y in y_test]

        y_train_class_rank = np.array(y_train_class_rank)
        y_test_class_rank = np.array(y_test_class_rank)

        list_classes_rank = [list_classes_copyright[class_rank], "not_" + list_classes_copyright[class_rank]]

        model = Classifier(
            model_name,
            architecture=architecture,
            list_classes=list_classes_rank,
            max_epoch=max_epoch,
            fold_number=fold_count,
            patience=patience,
            use_roc_auc=True,
            embeddings_name=embeddings_name,
            batch_size=batch_size,
            maxlen=maxlen,
            early_stop=early_stop,
            class_weights=class_weights,
            transformer_name=transformer,
        )

        if fold_count == 1:
            model.train(x_train, y_train_class_rank)
        else:
            model.train_nfold(x_train, y_train_class_rank)
        model.eval(x_test, y_test_class_rank)

        # saving the model
        # model.save()

    xtr, y_licenses = _read_data(
        "data/textClassification/licenses/copyrights-licenses-data-validated.csv", data_type="licenses"
    )

    # segment train and eval sets
    x_train, y_train, x_test, y_test = split_data_and_labels(xtr, y_licenses, 0.9)

    for class_rank in range(len(list_classes_licenses)):
        model_name = "licenses_" + list_classes_licenses[class_rank] + "_" + architecture
        class_weights = None

        batch_size, maxlen, patience, early_stop, max_epoch = configure(architecture)

        y_train_class_rank = [[1, 0] if y[class_rank] == 1.0 else [0, 1] for y in y_train]
        y_test_class_rank = [[1, 0] if y[class_rank] == 1.0 else [0, 1] for y in y_test]

        y_train_class_rank = np.array(y_train_class_rank)
        y_test_class_rank = np.array(y_test_class_rank)

        list_classes_rank = [list_classes_licenses[class_rank], "not_" + list_classes_licenses[class_rank]]

        model = Classifier(
            model_name,
            architecture=architecture,
            list_classes=list_classes_rank,
            max_epoch=max_epoch,
            fold_number=fold_count,
            patience=patience,
            use_roc_auc=True,
            embeddings_name=embeddings_name,
            batch_size=batch_size,
            maxlen=maxlen,
            early_stop=early_stop,
            class_weights=class_weights,
            transformer_name=transformer,
        )

        if fold_count == 1:
            model.train(x_train, y_train_class_rank)
        else:
            model.train_nfold(x_train, y_train_class_rank)
        model.eval(x_test, y_test_class_rank)

        # saving the model
        # model.save()


# classify a list of texts
def classify(texts, output_format, embeddings_name=None, architecture="gru", transformer=None):
    # load model
    model = Classifier("copyright_" + architecture)
    model.load()
    start_time = time.time()
    result = model.predict(texts, output_format)
    runtime = round(time.time() - start_time, 3)

    model = Classifier("license_" + architecture)
    model.load()
    result2 = model.predict(texts, output_format)

    runtime = round(time.time() - start_time, 3)
    if output_format == "json":
        result["runtime"] = runtime
    else:
        print("runtime: %s seconds " % (runtime))

    return result, result2


def _read_data(file_path, data_type="copyrights"):
    """
    Classification training data is expected in the following input format (csv or tsv):
    text_0    copyright_owner_0   license_0
    text_1    copyright_owner_1   license_1
    ...
    text_n    copyright_owner_n   license_n

    One of the value copyright_owner_i or license_i can be empty, in this case the training
    case is only relevant to one classifier.

    data_type indicates the data to retun according to the classifier, it is either
    "copyrights" or "licences".

    Note: final classification data include vectorized class values and the whole set is shuffled.
    """
    x, y = load_texts_and_classes_pandas_no_id(
        "data/textClassification/licenses/copyrights-licenses-data-validated.csv"
    )
    y_target = None
    list_classes = None
    if data_type == "copyrights":
        y_target = y[:, 0]
        list_classes = list_classes_copyright
    elif data_type == "licenses":
        y_target = y[:, 1]
        list_classes = list_classes_licenses
    else:
        print("invalid data type:", data_type)
        return None, None

    # remove rows when class valus is MISSINGVALUE
    def delete_rows(vecA, vecB, value):
        new_vecA = []
        new_vecB = []
        for i in range(len(vecB)):
            if vecB[i] != value and vecB[i] != "nan":
                new_vecA.append(vecA[i])
                new_vecB.append(vecB[i])
        return np.asarray(new_vecA), np.asarray(new_vecB)

    x, y_target = delete_rows(x, y_target, "MISSINGVALUE")

    xtr, y_target, _ = shuffle_triple_with_view(x, y_target)

    y_target_final = []
    for i in range(len(y_target)):
        index = list_classes.index(y_target[i])
        y_target_final.append(vectorizer(index, len(list_classes)))

    return xtr, np.array(y_target_final)


def report_training_copyrights(y):
    """
    Quick and dirty stats on the classes for copyrights owner
    """
    nb_publisher = 0
    nb_authors = 0
    nb_undecidable = 0
    for the_class in y:
        if the_class[0] == 1:
            nb_publisher += 1
        elif the_class[1] == 1:
            nb_authors += 1
        elif the_class[2] == 1:
            nb_undecidable += 1

    print("\ntotal context training cases:", len(y))

    print("\t- publisher owner:", nb_publisher)
    print("\t  not publisher owner:", str(len(y) - nb_publisher))

    print("\t- authors owner:", nb_authors)
    print("\t  not authors owner:", str(len(y) - nb_authors))

    print("\t- undecidable:", nb_undecidable)
    print("\t  not undecidable:", str(len(y) - nb_undecidable))


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Classify a copyright/license section using the DeLFT library")

    word_embeddings_examples = ["glove-840B", "fasttext-crawl", "word2vec"]
    pretrained_transformers_examples = ["bert-base-cased", "bert-large-cased", "allenai/scibert_scivocab_cased"]

    parser.add_argument("action")
    parser.add_argument("--fold-count", type=int, default=1)
    parser.add_argument(
        "--architecture", default="gru", help="type of model architecture to be used, one of " + str(architectures)
    )
    parser.add_argument(
        "--embedding",
        default=None,
        help="The desired pre-trained word embeddings using their descriptions in the file. "
        + "For local loading, use delft/resources-registry.json. "
        + "Be sure to use here the same name as in the registry, e.g. "
        + str(word_embeddings_examples)
        + " and that the path in the registry to the embedding file is correct on your system.",
    )
    parser.add_argument(
        "--transformer",
        default=None,
        help="The desired pre-trained transformer to be used in the selected architecture. "
        + "For local loading use, delft/resources-registry.json, and be sure to use here the same name as in the registry, e.g. "
        + str(pretrained_transformers_examples)
        + " and that the path in the registry to the model path is correct on your system. "
        + "HuggingFace transformers hub will be used otherwise to fetch the model, see https://huggingface.co/models "
        + "for model names",
    )

    args = parser.parse_args()

    if args.action not in ("train", "train_eval", "classify", "train_binary", "train_eval_binary"):
        print("action not specified, must be one of [train,train_binary,train_eval,train_eval_binary,classify]")

    embeddings_name = args.embedding
    transformer = args.transformer

    architecture = args.architecture
    if architecture not in architectures:
        print("unknown model architecture, must be one of " + str(architectures))

    if transformer is None and embeddings_name is None:
        # default word embeddings
        embeddings_name = "glove-840B"

    if args.action == "train":
        if args.fold_count < 1:
            raise ValueError("fold-count should be equal or more than 1")

        train(embeddings_name, args.fold_count, architecture=architecture, transformer=transformer)

    if args.action == "train_binary":
        if args.fold_count < 1:
            raise ValueError("fold-count should be equal or more than 1")

        train_binary(embeddings_name, args.fold_count, architecture=architecture, transformer=transformer)

    if args.action == "train_eval":
        if args.fold_count < 1:
            raise ValueError("fold-count should be equal or more than 1")

        y_test = train_and_eval(embeddings_name, args.fold_count, architecture=architecture, transformer=transformer)

    if args.action == "train_eval_binary":
        if args.fold_count < 1:
            raise ValueError("fold-count should be equal or more than 1")

        y_test = train_and_eval_binary(
            embeddings_name, args.fold_count, architecture=architecture, transformer=transformer
        )

    if args.action == "classify":
        someTexts = [
            "© 2005 Elsevier Inc. All rights reserved.",
            "This is an open-access article distributed under the terms of the Creative Commons Attribution License (http://creativecommons.org/licenses/by-nc/4.0), which permits unrestricted use, distribution, and reproduction in any medium, provided the original work is properly cited. Copyright © 2022 The Korean Association for Radiation Protection.",
            "© The Author(s) 2023.",
        ]
        result1, result2 = classify(
            someTexts, "json", architecture=architecture, embeddings_name=embeddings_name, transformer=transformer
        )
        print(json.dumps(result1, sort_keys=False, indent=4, ensure_ascii=False))
        print(json.dumps(result2, sort_keys=False, indent=4, ensure_ascii=False))



import datetime

import numpy as np
import tensorflow as tf
from packaging import version

from delft.sequenceLabelling.data_generator import DataGeneratorTransformers
from delft.sequenceLabelling.preprocess import Preprocessor
from delft.utilities.Tokenizer import tokenizeAndFilter


class Tagger(object):
    def __init__(
        self, model, model_config, embeddings=None, preprocessor: Preprocessor = None, transformer_preprocessor=None
    ):

        self.model = model
        self.preprocessor = preprocessor
        self.transformer_preprocessor = transformer_preprocessor
        self.model_config = model_config
        self.embeddings = embeddings

    def tag(self, texts, output_format, features=None, multi_gpu=False):
        if multi_gpu:
            strategy = tf.distribute.MirroredStrategy()
            print("Running with multi-gpu. Number of devices: {}".format(strategy.num_replicas_in_sync))

            # This trick avoid an exception being through when the --multi-gpu approach is used on a single GPU system.
            # It might be removed with TF 2.10 https://github.com/tensorflow/tensorflow/issues/50487
            if version.parse(tf.__version__) < version.parse("2.10.0"):
                import atexit

                atexit.register(strategy._extended._collective_ops._pool.close)  # type: ignore

            with strategy.scope():
                return self.tag_(texts, output_format, features)
        else:
            return self.tag_(texts, output_format, features)

    def tag_(self, texts, output_format, features=None):

        if output_format == "json":
            res = {
                "software": "DeLFT",
                "date": datetime.datetime.now().isoformat(),
                "model": self.model_config.model_name,
                "texts": [],
            }
        else:
            list_of_tags = []

        to_tokeniz = False
        if len(texts) > 0 and isinstance(texts[0], str):
            to_tokeniz = True

        # dirty fix warning! in the particular case of using tf-addons CRF layer and having a
        # single sequence in the input batch, a tensor shape error can happen in the CRF
        # viterbi_decoding loop. So to prevent this, we add a dummy second sequence in the batch
        # that we will remove after prediction
        dummy_case = False
        if self.model_config.use_crf and not self.model_config.use_chain_crf and len(texts) == 1:
            if features is None:
                if to_tokeniz:
                    texts.append(texts[0])
                else:
                    texts.append(["dummy"])
            else:
                texts.append(texts[0])
                # add a dummy feature vector for the token dummy...
                features.append(features[0])
            dummy_case = True
        # end of dirty fix

        generator = self.model.get_generator()
        predict_generator = generator(
            texts,
            None,
            batch_size=self.model_config.batch_size,
            preprocessor=self.preprocessor,
            bert_preprocessor=self.transformer_preprocessor,
            char_embed_size=self.model_config.char_embedding_size,
            max_sequence_length=self.model_config.max_sequence_length,
            embeddings=self.embeddings,
            tokenize=to_tokeniz,
            shuffle=False,
            features=features,
            output_input_offsets=True,
            use_chain_crf=self.model_config.use_chain_crf,
        )

        steps_done = 0
        steps = len(predict_generator)
        for generator_output in predict_generator:
            if dummy_case and steps_done == 1:
                break

            if steps_done == steps:
                break

            if isinstance(predict_generator, DataGeneratorTransformers):
                # the model uses transformer embeddings, so we need the input tokens to realign correctly the
                # labels and the inpit label texts

                # we need to remove one vector of the data corresponding to the marked tokens, this vector is not
                # expected by the model, but we need it to restore correctly the labels (which are produced
                # according to the sub-segmentation of wordpiece, not the expected segmentation)
                data = generator_output[0]

                input_offsets = data[-1]
                data = data[:-1]

                y_pred_batch = self.model.predict_on_batch(data)
                # y_pred_batch = np.argmax(y_pred_batch, -1)

                # results have been produced by a model using a transformer layer, so a few things to do
                # the labels are sparse, so integers and not one hot encoded
                # we need to restore back the labels for wordpiece to the labels for normal tokens
                # for this we can use the marked tokens provided by the generator
                new_y_pred_batch = []
                for y_pred_text, offsets_text in zip(y_pred_batch, input_offsets):
                    new_y_pred_text = []
                    # this is the result per sequence, realign labels:
                    for q in range(len(offsets_text)):
                        if offsets_text[q][0] == 0 and offsets_text[q][1] == 0:
                            # special token
                            continue
                        if offsets_text[q][0] != 0:
                            # added sub-token
                            continue
                        new_y_pred_text.append(y_pred_text[q])
                    new_y_pred_batch.append(new_y_pred_text)
                preds = new_y_pred_batch
            else:
                # no weirdness changes on the input
                preds = self.model.predict_on_batch(generator_output[0])

            for i in range(0, len(preds)):
                pred = [preds[i]]
                text = texts[i + (steps_done * self.model_config.batch_size)]

                if to_tokeniz:
                    tokens, offsets = tokenizeAndFilter(text)
                else:
                    # it is a list of string, so a string already tokenized
                    # note that in this case, offset are not present and json output is impossible
                    tokens = text
                    offsets = []

                if not self.model_config.use_crf or self.model_config.use_chain_crf:
                    tags = self._get_tags(pred)
                    prob = self._get_prob(pred)
                else:
                    tags = self._get_tags_sparse(pred)
                    prob = self._get_prob_sparse(pred)

                if output_format == "json":
                    piece = {}
                    piece["text"] = text
                    piece["entities"] = self._build_json_response(text, tokens, tags, prob, offsets)["entities"]
                    res["texts"].append(piece)
                else:
                    the_tags = list(zip(tokens, tags))
                    list_of_tags.append(the_tags)
            steps_done += 1

        if output_format == "json":
            return res
        else:
            return list_of_tags

    def _get_tags(self, pred):
        pred = np.argmax(pred, -1)
        tags = self.preprocessor.inverse_transform(pred[0])
        return tags

    def _get_tags_sparse(self, pred):
        tags = self.preprocessor.inverse_transform(pred[0])
        return tags

    def _get_prob(self, pred):
        prob = np.max(pred, -1)[0]
        return prob

    def _get_prob_sparse(self, pred):
        return [1.0] * len(pred[0])

    def _build_json_response(self, original_text, tokens, tags, prob, offsets):
        res = {"entities": []}
        chunks = get_entities_with_offsets(tags, offsets)
        for chunk_type, chunk_start, chunk_end, pos_start, pos_end in chunks:
            if prob is not None:
                score = float(np.average(prob[chunk_start:chunk_end]))
            else:
                score = 1.0

            entity_text = original_text[pos_start : pos_end + 1]
            entity = {
                "text": entity_text,
                "class": chunk_type,
                "score": score,
                "beginOffset": pos_start,
                "endOffset": pos_end,
            }
            res["entities"].append(entity)

        return res


def get_entities_with_offsets(seq, offsets):
    """
    Gets entities from sequence

    Args:
        seq (list): sequence of labels.
        offsets (list of integer pair): sequence of offset position

    Returns:
        list: list of (chunk_type, chunk_start, chunk_end, pos_start, pos_end)

    Example:
        >>> seq = ['B-PER', 'I-PER', 'O', 'B-LOC']
        >>> offsets = [(0,10), (11, 15), (16, 29), (30, 41)]
        >>> print(get_entities(seq))
        [('PER', 0, 2, 0, 15), ('LOC', 3, 4, 30, 41)]
    """

    i = 0
    chunks = []
    seq = seq + ["O"]  # add sentinel
    types = [tag.split("-")[-1] for tag in seq]
    max_length = min(len(seq) - 1, len(offsets))

    while i < max_length:
        if seq[i].startswith("B"):
            # if we are at the end of the offsets, we can stop immediately
            j = max_length
            if i + 1 != max_length:
                for j in range(i + 1, max_length + 1):
                    if seq[j].startswith("I") and types[j] == types[i]:
                        continue
                    break
            start_pos = offsets[i][0]
            end_pos = offsets[j - 1][1] - 1
            chunks.append((types[i], i, j, start_pos, end_pos))
            i = j
        else:
            i += 1
    return chunks



import math
import os

import numpy as np
from sklearn.metrics import log_loss, r2_score, roc_auc_score
from tf_keras.layers import (
    GRU,
    LSTM,
    Activation,
    Add,
    Bidirectional,
    Conv1D,
    Dense,
    Dropout,
    Flatten,
    GlobalAveragePooling1D,
    GlobalMaxPool1D,
    Input,
    MaxPooling1D,
    concatenate,
)
from tf_keras.models import Model
from tf_keras.optimizers import RMSprop
from transformers import create_optimizer

from delft.textClassification.data_generator import DataGenerator
from delft.utilities.Embeddings import load_resource_registry
from delft.utilities.misc import print_parameters
from delft.utilities.multiprocessing import get_multiprocessing_config
from delft.utilities.Transformer import DEFAULT_TRANSFORMER_TOKENIZER_DIR, TRANSFORMER_CONFIG_FILE_NAME, Transformer

architectures = [
    "lstm",
    "bidLstm_simple",
    "cnn",
    "cnn2",
    "cnn3",
    "gru_lstm",
    "lstm_cnn",
    "dpcnn",
    "gru",
    "gru_simple",
    "bert",
]


def getModel(model_config, training_config, load_pretrained_weights=True, local_path=None):
    """
    Return a model instance by its name. This is a facilitator function.
    """
    architecture = model_config.architecture

    # awww Python has no case/switch statement :D
    if architecture == "bidLstm_simple":
        model = bidLstm_simple(model_config, training_config)
    elif architecture == "lstm":
        model = lstm(model_config, training_config)
    elif architecture == "cnn":
        model = cnn(model_config, training_config)
    elif architecture == "cnn2":
        model = cnn2(model_config, training_config)
    elif architecture == "cnn3":
        model = cnn3(model_config, training_config)
    elif architecture == "lstm_cnn":
        model = lstm_cnn(model_config, training_config)
    elif architecture == "conv":
        model = dpcnn(model_config, training_config)
    elif architecture == "dpcnn":
        model = dpcnn(model_config, training_config)
    elif architecture == "gru":
        model = gru(model_config, training_config)
    elif architecture == "gru_lstm":
        model = gru_lstm(model_config, training_config)
    elif architecture == "gru_simple":
        model = gru_simple(model_config, training_config)
    elif architecture == "bert":
        model = bert(
            model_config, training_config, load_pretrained_weights=load_pretrained_weights, local_path=local_path
        )
    else:
        raise (OSError("The model type " + architecture + " is unknown"))
    return model


class BaseModel(object):
    """
    Base class for DeLFT text classification models

    Args:
        config (ModelConfig): DeLFT model configuration object
        ntags (integer): number of classes of the model
        load_pretrained_weights (boolean): used only when the model contains a transformer layer - indicate whether
                                           or not we load the pretrained weights of this transformer. For training
                                           a new model set it to True. When getting the full Keras model to load
                                           existing weights, set it False to avoid reloading the pretrained weights.
        local_path (string): used only when the model contains a transformer layer - the path where to load locally the
                             pretrained transformer. If None, the transformer model will be fetched from HuggingFace
                             transformers hub.
    """

    model = None
    parameters = {}
    registry = load_resource_registry("delft/resources-registry.json")
    transformer_config = None
    transformer_tokenizer = None

    def __init__(self, model_config, training_config, load_pretrained_weights=True, local_path=None):
        self.model_config = model_config
        self.training_config = training_config

    def update_parameters(self, model_config, training_config):
        for key in self.parameters:
            if hasattr(model_config, key):
                self.parameters[key] = getattr(model_config, key)
            elif hasattr(training_config, key):
                self.parameters[key] = getattr(training_config, key)

    def print_summary(self):
        if hasattr(self.model, "base_model"):
            self.model.base_model.summary()
        self.model.summary()

    def train_model(
        self,
        list_classes,
        batch_size,
        max_epoch,
        use_roc_auc,
        class_weights,
        training_generator,
        validation_generator,
        val_y,
        multiprocessing=True,
        patience=5,
        callbacks=None,
    ):

        nb_train_steps = (len(val_y) // batch_size) * max_epoch
        self.compile(nb_train_steps)

        best_loss = -1
        best_roc_auc = -1

        nb_workers, multiprocessing = get_multiprocessing_config(self.training_config, self.model_config)

        if validation_generator is None:
            # no early stop
            best_loss = self.model.fit(
                training_generator,
                use_multiprocessing=multiprocessing,
                workers=nb_workers,
                class_weight=class_weights,
                epochs=max_epoch,
                callbacks=callbacks,
            )
        else:
            best_weights = None
            current_epoch = 1
            best_epoch = 0
            while current_epoch <= max_epoch:
                loss = self.model.fit(
                    training_generator,
                    use_multiprocessing=multiprocessing,
                    workers=nb_workers,
                    class_weight=class_weights,
                    epochs=1,
                    callbacks=callbacks,
                )

                y_pred = self.model.predict(
                    validation_generator, use_multiprocessing=multiprocessing, workers=nb_workers
                )

                total_loss = 0.0
                total_roc_auc = 0.0

                # we distinguish 1-class and multiclass problems
                if len(list_classes) == 1:
                    total_loss = log_loss(val_y, y_pred, labels=[0, 1])
                    if len(np.unique(val_y)) == 1:
                        # roc_auc_score sklearn implementation is not working in this case, it needs more balanced batches
                        # a simple fix is to return the r2_score instead in this case (which is a regression score and not a loss)
                        roc_auc = r2_score(val_y, y_pred)
                        if roc_auc < 0:
                            roc_auc = 0
                    else:
                        total_roc_auc = roc_auc_score(val_y, y_pred)
                else:
                    for j in range(0, len(list_classes)):
                        loss = log_loss(val_y[:, j], y_pred[:, j], labels=[0, 1])
                        total_loss += loss
                        if len(np.unique(val_y[:, j])) == 1:
                            # roc_auc_score sklearn implementation is not working in this case, it needs more balanced batches
                            # a simple fix is to return the r2_score instead in this case (which is a regression score and not a loss)
                            roc_auc = r2_score(val_y[:, j], y_pred[:, j])
                            if roc_auc < 0:
                                roc_auc = 0
                        else:
                            roc_auc = roc_auc_score(val_y[:, j], y_pred[:, j])
                        total_roc_auc += roc_auc

                total_loss /= len(list_classes)
                total_roc_auc /= len(list_classes)
                if use_roc_auc:
                    print("Epoch {0} loss {1} best_loss {2} (for info) ".format(current_epoch, total_loss, best_loss))
                    print(
                        "Epoch {0} roc_auc {1} best_roc_auc {2} (for early stop) ".format(
                            current_epoch, total_roc_auc, best_roc_auc
                        )
                    )
                else:
                    print(
                        "Epoch {0} loss {1} best_loss {2} (for early stop) ".format(
                            current_epoch, total_loss, best_loss
                        )
                    )
                    print(
                        "Epoch {0} roc_auc {1} best_roc_auc {2} (for info) ".format(
                            current_epoch, total_roc_auc, best_roc_auc
                        )
                    )

                current_epoch += 1
                if total_loss < best_loss or best_loss == -1 or math.isnan(best_loss):
                    best_loss = total_loss
                    if not use_roc_auc:
                        best_weights = self.model.get_weights()
                        best_epoch = current_epoch
                elif not use_roc_auc:
                    if current_epoch - best_epoch == patience:
                        break

                if total_roc_auc > best_roc_auc or best_roc_auc == -1:
                    best_roc_auc = total_roc_auc
                    if use_roc_auc:
                        best_weights = self.model.get_weights()
                        best_epoch = current_epoch
                elif use_roc_auc:
                    if current_epoch - best_epoch == patience:
                        break

            self.model.set_weights(best_weights)

    def predict(self, predict_generator, use_main_thread_only=False):
        # default
        nb_workers = 6
        multiprocessing = True

        if use_main_thread_only:
            # worker at 0 means the training will be executed in the main thread
            nb_workers = 0
            multiprocessing = False

        y = self.model.predict(predict_generator, use_multiprocessing=multiprocessing, workers=nb_workers)
        return y

    def compile(self, train_size):
        # default compilation of the model.
        # train_size gives the number of steps for the traning, to be used for learning rate scheduler/decay
        self.model.compile(loss="binary_crossentropy", optimizer="adam", metrics=["accuracy"])

    def init_transformer(self, config, load_pretrained_weights=True, local_path=None):
        if config.transformer_name is None:
            # missing transformer name, no transformer layer to be initialized
            return None

        transformer = Transformer(config.transformer_name, resource_registry=self.registry, delft_local_path=local_path)
        print(config.transformer_name, "will be used, loaded via", transformer.loading_method)
        transformer_model = transformer.instantiate_layer(load_pretrained_weights=load_pretrained_weights)
        self.transformer_config = transformer.transformer_config
        transformer.init_preprocessor(max_sequence_length=config.maxlen)
        self.transformer_tokenizer = transformer.tokenizer

        return transformer_model

    def save(self, filepath):
        self.model.save_weights(filepath)

    def load(self, filepath):
        print("loading model weights", filepath)
        self.model.load_weights(filepath=filepath)


def train_folds(X, y, model_config, training_config, embeddings, models=None, callbacks=None):
    fold_count = model_config.fold_number
    max_epoch = training_config.max_epoch
    use_roc_auc = training_config.use_roc_auc
    class_weights = training_config.class_weights

    fold_size = len(X) // fold_count

    if models is None:
        models = []
        incremental = False
    else:
        incremental = True

    bert_data = False
    if model_config.transformer_name is not None:
        bert_data = True

    for fold_id in range(0, fold_count):
        fold_start = fold_size * fold_id
        fold_end = fold_start + fold_size

        if fold_id == fold_size - 1:
            fold_end = len(X)

        train_x = np.concatenate([X[:fold_start], X[fold_end:]])
        train_y = np.concatenate([y[:fold_start], y[fold_end:]])

        val_x = X[fold_start:fold_end]
        val_y = y[fold_start:fold_end]

        if incremental:
            foldModel = models[fold_id]
        else:
            foldModel = getModel(model_config, training_config)

        if fold_id == 0:
            print_parameters(model_config, training_config)
            foldModel.print_summary()

        print("\n------------------------ fold " + str(fold_id) + "--------------------------------------")

        training_generator = DataGenerator(
            train_x,
            train_y,
            batch_size=training_config.batch_size,
            maxlen=model_config.maxlen,
            list_classes=model_config.list_classes,
            embeddings=embeddings,
            bert_data=bert_data,
            shuffle=True,
            transformer_tokenizer=foldModel.transformer_tokenizer,
        )

        validation_generator = None
        if training_config.early_stop:
            validation_generator = DataGenerator(
                val_x,
                val_y,
                batch_size=training_config.batch_size,
                maxlen=model_config.maxlen,
                list_classes=model_config.list_classes,
                embeddings=embeddings,
                bert_data=bert_data,
                shuffle=False,
                transformer_tokenizer=foldModel.transformer_tokenizer,
            )

        foldModel.train_model(
            model_config.list_classes,
            training_config.batch_size,
            max_epoch,
            use_roc_auc,
            class_weights,
            training_generator,
            validation_generator,
            val_y,
            multiprocessing=training_config.multiprocessing,
            patience=training_config.patience,
            callbacks=callbacks,
        )

        if model_config.transformer_name is None:
            if incremental:
                models[fold_id] = foldModel
            else:
                models.append(foldModel)
        else:
            # if we are using a transformer layer in the architecture, we need to save the fold model on the disk
            directory = os.path.join("data/models/textClassification/", model_config.model_name)
            if not os.path.exists(directory):
                os.makedirs(directory)

            if fold_id == 0:
                if incremental:
                    models[0] = foldModel
                else:
                    models.append(foldModel)
                # save transformer config and tokenizer
                if foldModel.transformer_config is not None:
                    foldModel.transformer_config.to_json_file(os.path.join(directory, TRANSFORMER_CONFIG_FILE_NAME))
                if foldModel.transformer_tokenizer is not None:
                    foldModel.transformer_tokenizer.save_pretrained(
                        os.path.join(directory, DEFAULT_TRANSFORMER_TOKENIZER_DIR)
                    )

            model_path = os.path.join(directory, "model{0}_weights.hdf5".format(fold_id))
            foldModel.save(model_path)
            if fold_id != 0:
                del foldModel

    return models


def predict_folds(models, predict_generator, model_config, training_config, use_main_thread_only=False):
    fold_count = model_config.fold_number
    y_predicts_list = []
    for fold_id in range(0, fold_count):
        if model_config.transformer_name is not None:
            model = models[0]
            # load new weight from disk
            model_path = os.path.join(
                "data/models/textClassification/", model_config.model_name, "model{0}_weights.hdf5".format(fold_id)
            )
            model.load(model_path)
        else:
            model = models[fold_id]

        y_predicts = model.predict(predict_generator, use_main_thread_only=use_main_thread_only)
        y_predicts_list.append(y_predicts)

    y_predicts = np.ones(y_predicts_list[0].shape)
    for fold_predict in y_predicts_list:
        y_predicts *= fold_predict

    y_predicts **= 1.0 / len(y_predicts_list)

    return y_predicts


class lstm(BaseModel):
    """
    A Keras implementation of a LSTM classifier
    """

    name = "lstm"

    # default parameters
    parameters = {
        "max_features": 200000,
        "maxlen": 300,
        "embed_size": 300,
        "epoch": 40,
        "batch_size": 256,
        "dropout_rate": 0.3,
        "recurrent_dropout_rate": 0.3,
        "recurrent_units": 64,
        "dense_size": 32,
    }

    def __init__(self, model_config, training_config):
        super().__init__(model_config, training_config)
        self.update_parameters(model_config, training_config)
        nb_classes = len(model_config.list_classes)

        # basic LSTM
        input_layer = Input(
            shape=(self.parameters["maxlen"], self.parameters["embed_size"]),
        )
        x = LSTM(
            self.parameters["recurrent_units"],
            return_sequences=True,
            dropout=self.parameters["dropout_rate"],
            recurrent_dropout=self.parameters["dropout_rate"],
        )(input_layer)
        x = Dropout(self.parameters["dropout_rate"])(x)
        x_a = GlobalMaxPool1D()(x)
        x_b = GlobalAveragePooling1D()(x)
        x = concatenate([x_a, x_b])
        x = Dense(self.parameters["dense_size"], activation="relu")(x)
        x = Dropout(self.parameters["dropout_rate"])(x)
        x = Dense(nb_classes, activation="sigmoid")(x)
        self.model = Model(inputs=input_layer, outputs=x)


class bidLstm_simple(BaseModel):
    """
    A Keras implementation of a bidirectional LSTM  classifier
    """

    name = "bidLstm_simple"

    # default parameters
    parameters = {
        "max_features": 200000,
        "maxlen": 300,
        "embed_size": 300,
        "epoch": 25,
        "batch_size": 256,
        "dropout_rate": 0.3,
        "recurrent_dropout_rate": 0.3,
        "recurrent_units": 300,
        "dense_size": 256,
    }

    # bidirectional LSTM
    def __init__(self, model_config, training_config):
        super().__init__(model_config, training_config)
        self.update_parameters(model_config, training_config)
        nb_classes = len(model_config.list_classes)

        input_layer = Input(
            shape=(self.parameters["maxlen"], self.parameters["embed_size"]),
        )
        x = Bidirectional(
            LSTM(
                self.parameters["recurrent_units"],
                return_sequences=True,
                dropout=self.parameters["dropout_rate"],
                recurrent_dropout=self.parameters["dropout_rate"],
            )
        )(input_layer)
        x = Dropout(self.parameters["dropout_rate"])(x)
        x_a = GlobalMaxPool1D()(x)
        x_b = GlobalAveragePooling1D()(x)
        x = concatenate([x_a, x_b])
        x = Dense(self.parameters["dense_size"], activation="relu")(x)
        x = Dropout(self.parameters["dropout_rate"])(x)
        x = Dense(nb_classes, activation="sigmoid")(x)
        self.model = Model(inputs=input_layer, outputs=x)


class cnn(BaseModel):
    """
    A Keras implementation of a CNN classifier
    """

    name = "cnn"

    # default parameters
    parameters = {
        "max_features": 200000,
        "maxlen": 250,
        "embed_size": 300,
        "epoch": 30,
        "batch_size": 256,
        "dropout_rate": 0.3,
        "recurrent_dropout_rate": 0.3,
        "recurrent_units": 64,
        "dense_size": 32,
    }

    # conv+GRU with embeddings
    def __init__(self, model_config, training_config):
        super().__init__(model_config, training_config)
        self.update_parameters(model_config, training_config)
        nb_classes = len(model_config.list_classes)

        input_layer = Input(
            shape=(self.parameters["maxlen"], self.parameters["embed_size"]),
        )
        x = Dropout(self.parameters["dropout_rate"])(input_layer)
        x = Conv1D(filters=self.parameters["recurrent_units"], kernel_size=2, padding="same", activation="relu")(x)
        x = MaxPooling1D(pool_size=2)(x)
        x = Conv1D(filters=self.parameters["recurrent_units"], kernel_size=2, padding="same", activation="relu")(x)
        x = MaxPooling1D(pool_size=2)(x)
        x = Conv1D(filters=self.parameters["recurrent_units"], kernel_size=2, padding="same", activation="relu")(x)
        x = MaxPooling1D(pool_size=2)(x)
        x = GRU(self.parameters["recurrent_units"])(x)
        x = Dropout(self.parameters["dropout_rate"])(x)
        x = Dense(self.parameters["dense_size"], activation="relu")(x)
        x = Dense(nb_classes, activation="sigmoid")(x)
        self.model = Model(inputs=input_layer, outputs=x)


class cnn2(BaseModel):
    """
    A Keras implementation of a CNN classifier (variant)
    """

    name = "cnn2"

    # default parameters
    parameters = {
        "max_features": 200000,
        "maxlen": 250,
        "embed_size": 300,
        "epoch": 30,
        "batch_size": 256,
        "dropout_rate": 0.3,
        "recurrent_dropout_rate": 0.3,
        "recurrent_units": 64,
        "dense_size": 32,
    }

    def __init__(self, model_config, training_config):
        super().__init__(model_config, training_config)
        self.update_parameters(model_config, training_config)
        nb_classes = len(model_config.list_classes)

        input_layer = Input(
            shape=(self.parameters["maxlen"], self.parameters["embed_size"]),
        )
        x = Dropout(self.parameters["dropout_rate"])(input_layer)
        x = Conv1D(filters=self.parameters["recurrent_units"], kernel_size=2, padding="same", activation="relu")(x)
        x = Conv1D(filters=self.parameters["recurrent_units"], kernel_size=2, padding="same", activation="relu")(x)
        x = Conv1D(filters=self.parameters["recurrent_units"], kernel_size=2, padding="same", activation="relu")(x)
        x = GRU(
            self.parameters["recurrent_units"],
            return_sequences=False,
            dropout=self.parameters["dropout_rate"],
            recurrent_dropout=self.parameters["dropout_rate"],
        )(x)
        x = Dense(self.parameters["dense_size"], activation="relu")(x)
        x = Dense(nb_classes, activation="sigmoid")(x)
        self.model = Model(inputs=input_layer, outputs=x)


class cnn3(BaseModel):
    """
    A Keras implementation of a CNN classifier (variant)
    """

    name = "cnn3"

    # default parameters
    parameters = {
        "max_features": 200000,
        "maxlen": 300,
        "embed_size": 300,
        "epoch": 30,
        "batch_size": 256,
        "dropout_rate": 0.3,
        "recurrent_dropout_rate": 0.3,
        "recurrent_units": 64,
        "dense_size": 32,
    }

    def __init__(self, model_config, training_config):
        super().__init__(model_config, training_config)
        self.update_parameters(model_config, training_config)
        nb_classes = len(model_config.list_classes)

        input_layer = Input(
            shape=(self.parameters["maxlen"], self.parameters["embed_size"]),
        )
        x = GRU(
            self.parameters["recurrent_units"],
            return_sequences=True,
            dropout=self.parameters["dropout_rate"],
            recurrent_dropout=self.parameters["dropout_rate"],
        )(input_layer)
        x = Conv1D(filters=self.parameters["recurrent_units"], kernel_size=2, padding="same", activation="relu")(x)
        x = MaxPooling1D(pool_size=2)(x)
        x = Conv1D(filters=self.parameters["recurrent_units"], kernel_size=2, padding="same", activation="relu")(x)
        x = MaxPooling1D(pool_size=2)(x)
        x = Conv1D(filters=self.parameters["recurrent_units"], kernel_size=2, padding="same", activation="relu")(x)
        x = MaxPooling1D(pool_size=2)(x)
        x_a = GlobalMaxPool1D()(x)
        x_b = GlobalAveragePooling1D()(x)
        x = concatenate([x_a, x_b])
        x = Dense(self.parameters["dense_size"], activation="relu")(x)
        x = Dense(nb_classes, activation="sigmoid")(x)
        self.model = Model(inputs=input_layer, outputs=x)


class lstm_cnn(BaseModel):
    """
    A Keras implementation of a LSTM + CNN classifier
    """

    name = "lstm_cnn"

    # default parameters
    parameters = {
        "max_features": 200000,
        "maxlen": 250,
        "embed_size": 300,
        "epoch": 30,
        "batch_size": 256,
        "dropout_rate": 0.3,
        "recurrent_dropout_rate": 0.3,
        "recurrent_units": 64,
        "dense_size": 32,
    }

    # LSTM + conv
    def __init__(self, model_config, training_config):
        super().__init__(model_config, training_config)
        self.update_parameters(model_config, training_config)
        nb_classes = len(model_config.list_classes)

        input_layer = Input(
            shape=(self.parameters["maxlen"], self.parameters["embed_size"]),
        )
        x = LSTM(
            self.parameters["recurrent_units"],
            return_sequences=True,
            dropout=self.parameters["dropout_rate"],
            recurrent_dropout=self.parameters["dropout_rate"],
        )(input_layer)
        x = Dropout(self.parameters["dropout_rate"])(x)

        x = Conv1D(filters=self.parameters["recurrent_units"], kernel_size=2, padding="same", activation="relu")(x)
        x = Conv1D(filters=300, kernel_size=5, padding="valid", activation="tanh", strides=1)(x)
        x_a = GlobalMaxPool1D()(x)
        x_b = GlobalAveragePooling1D()(x)
        x = concatenate([x_a, x_b])
        x = Dense(self.parameters["dense_size"], activation="relu")(x)
        x = Dropout(self.parameters["dropout_rate"])(x)
        x = Dense(nb_classes, activation="sigmoid")(x)
        self.model = Model(inputs=input_layer, outputs=x)


class gru(BaseModel):
    """
    A Keras implementation of a Bidirectional GRU classifier
    """

    name = "gru"

    # default parameters
    parameters = {
        "max_features": 200000,
        "maxlen": 300,
        "embed_size": 300,
        "epoch": 30,
        "batch_size": 256,
        "dropout_rate": 0.3,
        "recurrent_dropout_rate": 0.3,
        "recurrent_units": 64,
        "dense_size": 32,
    }

    # 2 bid. GRU
    def __init__(self, model_config, training_config):
        super().__init__(model_config, training_config)
        self.update_parameters(model_config, training_config)
        nb_classes = len(model_config.list_classes)

        input_layer = Input(
            shape=(self.parameters["maxlen"], self.parameters["embed_size"]),
        )
        x = Bidirectional(
            GRU(
                self.parameters["recurrent_units"],
                return_sequences=True,
                dropout=self.parameters["dropout_rate"],
                recurrent_dropout=self.parameters["recurrent_dropout_rate"],
            )
        )(input_layer)
        x = Dropout(self.parameters["dropout_rate"])(x)
        x = Bidirectional(
            GRU(
                self.parameters["recurrent_units"],
                return_sequences=True,
                dropout=self.parameters["dropout_rate"],
                recurrent_dropout=self.parameters["recurrent_dropout_rate"],
            )
        )(x)
        x_a = GlobalMaxPool1D()(x)
        x_b = GlobalAveragePooling1D()(x)
        x = concatenate([x_a, x_b], axis=1)
        x = Dense(self.parameters["dense_size"], activation="relu")(x)
        output_layer = Dense(nb_classes, activation="sigmoid")(x)
        self.model = Model(inputs=input_layer, outputs=output_layer)

    def compile(self, train_size):
        self.model.compile(loss="binary_crossentropy", optimizer=RMSprop(global_clipnorm=1), metrics=["accuracy"])


class gru_simple(BaseModel):
    """
    A Keras implementation of a one layer Bidirectional GRU classifier
    """

    name = "gru_simple"

    # default parameters
    parameters = {
        "max_features": 200000,
        "maxlen": 300,
        "embed_size": 300,
        "epoch": 30,
        "batch_size": 512,
        "dropout_rate": 0.3,
        "recurrent_dropout_rate": 0.3,
        "recurrent_units": 64,
        "dense_size": 32,
    }

    # 1 layer bid GRU
    def __init__(self, model_config, training_config):
        super().__init__(model_config, training_config)
        self.update_parameters(model_config, training_config)
        nb_classes = len(model_config.list_classes)

        input_layer = Input(
            shape=(self.parameters["maxlen"], self.parameters["embed_size"]),
        )
        x = Bidirectional(
            GRU(
                self.parameters["recurrent_units"],
                return_sequences=True,
                dropout=self.parameters["dropout_rate"],
                recurrent_dropout=self.parameters["dropout_rate"],
            )
        )(input_layer)
        x_a = GlobalMaxPool1D()(x)
        x_b = GlobalAveragePooling1D()(x)
        x = concatenate([x_a, x_b], axis=1)
        x = Dense(self.parameters["dense_size"], activation="relu")(x)
        output_layer = Dense(nb_classes, activation="sigmoid")(x)
        self.model = Model(inputs=input_layer, outputs=output_layer)

    def compile(self, train_size):
        self.model.compile(loss="binary_crossentropy", optimizer=RMSprop(global_clipnorm=1), metrics=["accuracy"])


class gru_lstm(BaseModel):
    """
    A Keras implementation of a mixed Bidirectional GRU and LSTM classifier
    """

    name = "gru_lstm"

    # default parameters
    parameters = {
        "max_features": 200000,
        "maxlen": 300,
        "embed_size": 300,
        "epoch": 30,
        "batch_size": 256,
        "dropout_rate": 0.3,
        "recurrent_dropout_rate": 0.3,
        "recurrent_units": 64,
        "dense_size": 32,
    }

    # bid GRU + bid LSTM
    def __init__(self, model_config, training_config):
        super().__init__(model_config, training_config)
        self.update_parameters(model_config, training_config)
        nb_classes = len(model_config.list_classes)

        input_layer = Input(
            shape=(self.parameters["maxlen"], self.parameters["embed_size"]),
        )
        x = Bidirectional(
            GRU(
                self.parameters["recurrent_units"],
                return_sequences=True,
                dropout=self.parameters["dropout_rate"],
                recurrent_dropout=self.parameters["recurrent_dropout_rate"],
            )
        )(input_layer)
        x = Dropout(self.parameters["dropout_rate"])(x)
        x = Bidirectional(
            LSTM(
                self.parameters["recurrent_units"],
                return_sequences=True,
                dropout=self.parameters["dropout_rate"],
                recurrent_dropout=self.parameters["recurrent_dropout_rate"],
            )
        )(x)
        x_a = GlobalMaxPool1D()(x)
        x_b = GlobalAveragePooling1D()(x)
        x = concatenate([x_a, x_b])
        x = Dense(self.parameters["dense_size"], activation="relu")(x)
        output_layer = Dense(nb_classes, activation="sigmoid")(x)
        self.model = Model(inputs=input_layer, outputs=output_layer)

    def compile(self, train_size):
        self.model.compile(loss="binary_crossentropy", optimizer=RMSprop(global_clipnorm=1), metrics=["accuracy"])


class dpcnn(BaseModel):
    """
    A Keras implementation of a DPCNN classifier
    """

    name = "dpcnn"

    # default parameters
    parameters = {
        "max_features": 200000,
        "maxlen": 300,
        "embed_size": 300,
        "epoch": 25,
        "batch_size": 256,
        "dropout_rate": 0.3,
        "recurrent_dropout_rate": 0.3,
        "recurrent_units": 64,
        "dense_size": 32,
    }

    # DPCNN
    def __init__(self, model_config, training_config):
        super().__init__(model_config, training_config)
        self.update_parameters(model_config, training_config)
        nb_classes = len(model_config.list_classes)

        input_layer = Input(
            shape=(self.parameters["maxlen"], self.parameters["embed_size"]),
        )
        # first block
        X_shortcut1 = input_layer
        X = Conv1D(filters=self.parameters["recurrent_units"], kernel_size=2, strides=3)(X_shortcut1)
        X = Activation("relu")(X)
        X = Conv1D(filters=self.parameters["recurrent_units"], kernel_size=2, strides=3)(X)
        X = Activation("relu")(X)

        # connect shortcut to the main path
        X = Activation("relu")(X_shortcut1)  # pre activation
        X = Add()([X_shortcut1, X])
        X = MaxPooling1D(pool_size=3, strides=2, padding="valid")(X)

        # second block
        X_shortcut2 = X
        X = Conv1D(filters=self.parameters["recurrent_units"], kernel_size=2, strides=3)(X)
        X = Activation("relu")(X)
        X = Conv1D(filters=self.parameters["recurrent_units"], kernel_size=2, strides=3)(X)
        X = Activation("relu")(X)

        # connect shortcut to the main path
        X = Activation("relu")(X_shortcut2)  # pre activation
        X = Add()([X_shortcut2, X])
        X = MaxPooling1D(pool_size=2, strides=2, padding="valid")(X)

        # Output
        X = Flatten()(X)
        X = Dense(nb_classes, activation="sigmoid")(X)

        self.model = Model(inputs=input_layer, outputs=X, name="dpcnn")


class bert(BaseModel):
    """
    A Keras implementation of a BERT classifier for fine-tuning, with BERT layer to be
    instanciated with a pre-trained BERT model
    """

    name = "bert"
    bert_config = None

    # default parameters
    parameters = {"dense_size": 512, "max_seq_len": 512, "dropout_rate": 0.1, "batch_size": 10}

    # simple BERT classifier with TF transformers, architecture equivalent to the original BERT implementation
    def __init__(self, model_config, training_config, load_pretrained_weights=True, local_path=None):
        super().__init__(model_config, training_config, load_pretrained_weights, local_path)
        self.update_parameters(model_config, training_config)
        nb_classes = len(model_config.list_classes)

        transformer_model = self.init_transformer(
            model_config, load_pretrained_weights=load_pretrained_weights, local_path=local_path
        )

        input_ids_in = Input(shape=(None,), name="input_token", dtype="int32")
        # input_masks_in = Input(shape=(None,), name='masked_token', dtype='int32')
        # embedding_layer = transformer_model(input_ids_in, attention_mask=input_masks_in)[1]

        # get the pooler_output as in the original BERT implementation
        embedding_layer = transformer_model(input_ids_in)[1]
        cls_out = Dropout(self.parameters["dropout_rate"])(embedding_layer)
        logits = Dense(units=nb_classes, activation="softmax")(cls_out)

        self.model = Model(inputs=[input_ids_in], outputs=logits)
        # self.model = Model(inputs=[input_ids_in, input_masks_in], outputs=logits)

    def compile(self, train_size):
        # optimizer = Adam(learning_rate=2e-5, clipnorm=1)
        optimizer, lr_schedule = create_optimizer(
            init_lr=self.training_config.learning_rate,
            num_train_steps=train_size,
            weight_decay_rate=0.01,
            num_warmup_steps=0.1 * train_size,
        )
        self.model.compile(optimizer=optimizer, loss="binary_crossentropy", metrics=["accuracy"])



"""
Utility functions for multiprocessing configuration in training.
"""


def get_multiprocessing_config(training_config, model_config):
    """
    Determine the number of workers and multiprocessing mode for training.

    Args:
        training_config: TrainingConfig object with num_workers and multiprocessing settings
        model_config: ModelConfig object with transformer_name

    Returns:
        tuple: (nb_workers, use_multiprocessing)
    """
    multiprocessing = True

    if hasattr(training_config, "num_workers") and training_config.num_workers is not None:
        nb_workers = training_config.num_workers
        if nb_workers == 0:
            nb_workers = 1
            multiprocessing = False
        else:
            multiprocessing = training_config.multiprocessing and nb_workers > 1
    else:
        nb_workers = 6
        multiprocessing = training_config.multiprocessing

    if model_config.transformer_name is not None:
        nb_workers = 1
        multiprocessing = False

    return nb_workers, multiprocessing



import json


# Model parameter
class ModelConfig(object):
    def __init__(
        self,
        model_name="",
        architecture="gru",
        embeddings_name="glove-840B",
        list_classes=[],
        char_emb_size=0,
        word_emb_size=300,
        dropout=0.5,
        recurrent_dropout=0.25,
        use_char_feature=False,
        maxlen=300,
        fold_number=1,
        batch_size=64,
        dense_size=32,
        transformer_name=None,
    ):

        self.model_name = model_name
        self.architecture = architecture
        self.embeddings_name = embeddings_name

        # self.vocab_size = None
        # self.char_vocab_size = None

        self.char_embedding_size = char_emb_size
        self.word_embedding_size = word_emb_size
        self.dropout = dropout
        self.recurrent_dropout = recurrent_dropout
        self.maxlen = maxlen
        self.dense_size = dense_size

        self.use_char_feature = use_char_feature
        self.list_classes = list_classes
        self.fold_number = fold_number
        self.batch_size = batch_size  # this is the batch size for test and prediction

        self.transformer_name = transformer_name

    def save(self, file):
        with open(file, "w") as f:
            json.dump(vars(self), f, sort_keys=False, indent=4)

    @classmethod
    def load(cls, file):
        with open(file) as f:
            variables = json.load(f)
            self = cls()
            for key, val in variables.items():
                setattr(self, key, val)
        return self


# Training parameter
class TrainingConfig(object):
    def __init__(
        self,
        learning_rate,
        batch_size=256,
        optimizer="adam",
        lr_decay=0.9,
        clip_gradients=5.0,
        max_epoch=50,
        patience=5,
        early_stop=True,
        use_roc_auc=True,
        class_weights=None,
        multiprocessing=True,
        num_workers=1,
    ):

        self.batch_size = batch_size  # this is the batch size for training
        self.optimizer = optimizer
        self.learning_rate = learning_rate
        self.lr_decay = lr_decay
        self.clip_gradients = clip_gradients
        self.max_epoch = max_epoch
        self.patience = patience
        self.use_roc_auc = use_roc_auc
        self.class_weights = class_weights
        self.multiprocessing = multiprocessing
        self.num_workers = num_workers
        self.early_stop = early_stop
